import numpy as np
from matplotlib import pyplot as plt
from typing import List
from src import hidr

class submercado(object):

    # Classe contendo informacoes sobre os arquivos de entrada e saida
    Codigo = None
    Nome = None
    Mercado = None
    NaoSimuladas = None
    Ficticio = None

    def __init__(self, nanos):
        self.Mercado = np.zeros((nanos+1,12), 'd')
        self.NaoSimuladas = np.zeros((nanos,12), 'd')
        self.CustoDeficit = np.zeros(4, 'd')
        self.ProfundidadeDeficit = np.zeros(4, 'd')

    # Parametros Temporais
    EArmMax = None              # Energia Armazenavel Maxima (iano,imes)
    ENA = None
    EFIO = None
    EC = None

    def CalcEArmMax(self, usinas:List[hidr]):
        self.EArmMax = np.zeros( (5,12), 'd')
        for iano in range(5):
            for imes in range(12):
                self.EArmMax[iano][imes] = 0.
                for iusina in usinas:
                    if iusina.VolUtil > 0 and iusina.Sist == self.Codigo and iusina.StatusVolMorto[iano][imes] == 2:
                        self.EArmMax[iano][imes] = self.EArmMax[iano][imes] + iusina.RoAcum[iano][imes]*iusina.VolUtil/2.63


    def CalcENA (self, usinas:List[hidr]):

        anos_hist = len(usinas[0].Vazoes)

        self.ENA = np.zeros( (5,12,anos_hist), 'd')
        self.EC = np.zeros( (5,12,anos_hist), 'd')
        self.EFIO = np.zeros( (5,12,anos_hist), 'd')

        for iano in range(5):
            for imes in range(12):
                for iusina in usinas:
                    if iusina.Sist == self.Codigo and iusina.StatusVolMorto[iano][imes] == 2:
                        if iusina.VolUtil > 0:
                            self.EC[iano,imes] = self.EC[iano,imes] + iusina.RoAcum[iano][imes]*iusina.QIncEntreRes(usinas, iano, imes)
                        else:
                            if iusina.StatusMotoriz[iano][imes] == 2:
                                self.EFIO[iano,imes] = self.EFIO[iano,imes] +  iusina.RoEquiv[iano][imes]*iusina.QIncEntreRes(usinas, iano, imes)

        self.ENA = self.EC + self.EFIO


    #############################################################################
    #############################################################################
    # Graficos Diversos
    #############################################################################
    #############################################################################

    def PlotaEArmMax(self, usinas:List[hidr]):

        f, (ax) = plt.subplots(1, 1)

        nr_lin = len(self.EArmMax) - 1
        nr_col = len(self.EArmMax[0])
        total = np.zeros((nr_lin, nr_col), 'd')
        nanos = nr_lin

        y = self.EArmMax[0:nr_lin][0:nr_col]
        FioJusante  = np.zeros ( (nanos,12) , 'd')
        ContJusante = np.zeros ( (nanos,12) , 'd')
        for iano in range(nanos):
            for imes in range(12):
                for iusina in usinas:
                    if iusina.VolUtil > 0 and iusina.Sist == self.Codigo and iusina.StatusVolMorto[iano][imes] == 2:
                        FioJusante[iano][imes] += iusina.RoAcum_B_Sist[iano][imes]* iusina.VolUtil / 2.63
                        ContJusante[iano][imes] += iusina.RoAcum_C_Sist[iano][imes]* iusina.VolUtil / 2.63

        cor = 'black'
        linha = '-'
        LineWidth = 3
        Texto = "EARMAX - Avg: " + str(np.trunc(np.mean(y)))
        ax.plot(np.arange(1, nr_lin * nr_col + 1), y.reshape(nr_lin * nr_col, ), linha, color=cor,
                lw=LineWidth, label=Texto)
        ax.fill_between(np.arange(1, nr_lin * nr_col + 1), (FioJusante+ContJusante).reshape(nr_lin * nr_col, ),
                        y.reshape(nr_lin * nr_col, ), facecolor=cor, alpha=0.1)

        cor = 'blue'
        linha = '--'
        LineWidth = 2
        Texto = "Parc Fio Jus - Avg: " + str(np.trunc(np.mean(FioJusante)))
        ax.plot(np.arange(1, nr_lin * nr_col + 1), FioJusante.reshape(nr_lin * nr_col, ), linha, color=cor,
                lw=LineWidth, label=Texto)
        ax.fill_between(np.arange(1, nr_lin * nr_col + 1), 0,
                        y.reshape(nr_lin * nr_col, ), facecolor=cor, alpha=0.1)

        cor = 'chocolate'
        linha = 'x'
        LineWidth = 2
        Texto = "Parc Cont Jus - Avg: " + str(np.trunc(np.mean(ContJusante)))
        ax.plot(np.arange(1, nr_lin * nr_col + 1), (FioJusante+ContJusante).reshape(nr_lin * nr_col, ),
                linha, color=cor,lw=LineWidth, label=Texto)
        ax.fill_between(np.arange(1, nr_lin * nr_col + 1), FioJusante.reshape(nr_lin * nr_col, ),
                        ContJusante.reshape(nr_lin * nr_col, ), facecolor=cor, alpha=0.1)

        titulo = 'Evolucao da Energia Armazenavel Maxima \n do Submercado ' + self.Nome
        f.canvas.set_window_title(titulo)

        ax.set_title(titulo, fontsize=16)
        ax.set_xlabel('Meses de Estudo', fontsize=14)
        ax.set_ylabel('Energia Armaz. Maxima (MWmes)', fontsize=14)
        ax.legend(fontsize=12)

        plt.show()

    def PlotaMercado(self):

        f, (ax) = plt.subplots(1, 1)

        nr_lin = len(self.Mercado)-1
        nr_col = len(self.Mercado[0])

        if self.Ficticio == 0:
            carga = self.Mercado[0:nr_lin][0:nr_col]
            pq = self.NaoSimuladas
            ax.plot(np.arange(1, nr_lin*nr_col+1) , carga.reshape( nr_lin*nr_col, ), 'r-',lw=3, label = 'Mercado Total')
            ax.plot(np.arange(1, nr_lin*nr_col+1) , pq.reshape( nr_lin*nr_col, ), 'g--',lw=2, label = 'Usinas Nao Simuladas')
            ax.fill_between(np.arange(1, nr_lin*nr_col+1), 0, pq.reshape( nr_lin*nr_col, ), facecolor='g', alpha=0.1)
            ax.fill_between(np.arange(1, nr_lin*nr_col+1), pq.reshape( nr_lin*nr_col, ), carga.reshape( nr_lin*nr_col, ), facecolor='r', alpha=0.1)

        titulo = 'Evolucao do Mercado Total do ' + self.Nome
        f.canvas.set_window_title(titulo)

        ax.set_title(titulo,  fontsize = 16)
        ax.set_xlabel('Meses de Estudo', fontsize = 14)
        ax.set_ylabel('Demanda de Energia (MWmes)', fontsize = 14)
        ax.legend(fontsize = 12)

        plt.show()

    def PlotaENA(self):

        nanos = len(self.ENA)
        nmeses = len(self.ENA[0])
        nseries = len(self.ENA[0][0])

        f, (ax1, ax2, ax3) = plt.subplots(3, 1, sharex=True)

        eixo_x = np.arange(1,nanos*nmeses+1)

        ENAS = np.zeros( (nseries,nanos*nmeses), 'd' )
        ECS = np.zeros( (nseries,nanos*nmeses), 'd' )
        EFIOS = np.zeros( (nseries,nanos*nmeses), 'd' )

        contador = 0
        for iano in range(nanos):
            for imes in range(nmeses):
                for iserie in range(nseries):
                    ENAS[iserie][contador] = self.ENA[iano][imes][iserie]
                    ECS[iserie][contador] = self.EC[iano][imes][iserie]
                    EFIOS[iserie][contador] = self.EFIO[iano][imes][iserie]
                contador = contador + 1

        ax1.plot(eixo_x,ENAS.transpose(),'c-')
        ax2.plot(eixo_x,ECS.transpose(),'c-')
        ax3.plot(eixo_x,EFIOS.transpose(),'c-')

        mediaENA = np.mean(ENAS, axis=0)
        mediaEC = np.mean(ECS, axis=0)
        mediaEFIO = np.mean(EFIOS, axis=0)

        ax1.plot(eixo_x,mediaENA,'r-',lw=3)
        ax2.plot(eixo_x,mediaEC,'r-',lw=3)
        ax3.plot(eixo_x,mediaEFIO,'r-',lw=3)

        desvioENA = np.nanstd(ENAS, axis=0)
        desvioEC = np.nanstd(ECS, axis=0)
        desvioEFIO = np.nanstd(EFIOS, axis=0)

        ax1.plot(eixo_x,mediaENA+desvioENA,'r-.',lw=2)
        ax1.plot(eixo_x,mediaENA-desvioENA,'r-.',lw=2)
        ax2.plot(eixo_x,mediaEC+desvioEC,'r-.',lw=2)
        ax2.plot(eixo_x,mediaEC-desvioEC,'r-.',lw=2)
        ax3.plot(eixo_x,mediaEFIO+desvioEFIO,'r-.',lw=2)
        ax3.plot(eixo_x,mediaEFIO-desvioEFIO,'r-.',lw=2)

        ax1.set_xticks(np.arange(1,60))
        ax1.set_ylabel('ENA (MWmes)')
        ax2.set_ylabel('EC (MWmes)')
        ax3.set_ylabel('EFIO (MWMes)')
        tituloENA =  'ENA do Submercado ' + self.Nome
        tituloEC = 'EC do Submercado ' + self.Nome
        tituloEFIO = 'EFIO do Submercado ' + self.Nome
        ax1.set_title(tituloENA, fontsize = 13)
        ax2.set_title(tituloEC, fontsize =13)
        ax3.set_title(tituloEFIO, fontsize=13)

        plt.show()





class term(object):
    # Dados de cadastro das usinas termeletrica (presentes no TERM.DAT)
    Codigo = None
    Nome = None
    Potencia = None
    FCMax = None
    TEIF = None
    IP = None
    GTMin = None

    # Dados Adicionais Especificados no arquivo de configuracao termica (CONFT)
    Sist = None
    Status = None
    Classe = None

    # Dados Adicionais Especificados no arquivo de classe termica (CLAST)
    Custo = None
    NomeClasse = None
    TipoComb = None

    def insere(self, custo, gmax):
        self.custo = custo
        self.gmax = gmax



from src import fcf
from itertools import product
from cvxopt.modeling import op, variable, matrix
import cvxopt.solvers as solvers
# import matplotlib
# matplotlib.use('Agg')
from matplotlib import pyplot as plt
import numpy as np
from multiprocessing import Pool
from src import pmo
from src import dadosgerais
from functools import partial
from timeit import default_timer as timer


class mddh(object):
    # Classe contendo informacoes sobre os arquivos de entrada e saida
    caso = None
    dger = None
    npmc = 1

    # Define listas
    cadr_uh = []  # Lista com usinas hidraulicas do cadastro (HIDR.DAT)
    cadr_ut = []  # Lista com usinas termicas do cadastro (TERM.DAT)
    conf_ut = []  # Lista com usinas termicas da configuracao em estudo (CONFT.DAT)
    conf_uh = []  # Lista com usinas hidraulicas da configuracao em estudo (CONFHD.DAT)
    cfuturo = []  # Lista com funcoes de custo futuro obtidas pela PDE ou PDDE
    ree = []  # Lista com os reservatorios equivalentes de energia
    submercado = []  # Lista com os submercados
    intercambio = []  # Lista com intercambios entre submercados

    # Parametros Gerais
    ano_ini = 2018
    mes_ini = 12
    nanos = 5
    cdef = 4000
    mercado = 270
    ordmaxparp = 6

    def __init__(self, diretorio):

        # Leitura de arquivos
        self.caso = pmo(diretorio)
        self.caso.le_caso()

        self.cadr_uh = self.caso.le_hidr(self.cadr_uh)
        self.conf_uh = self.caso.le_confh(self.conf_uh, self.cadr_uh, self.nanos)
        self.cadr_ut = self.caso.le_term(self.cadr_ut)
        self.conf_ut = self.caso.le_conft(self.conf_ut, self.cadr_ut)
        self.conf_ut = self.caso.le_clast(self.conf_ut)
        self.conf_uh = self.caso.le_modif(self.conf_uh, self.ano_ini, self.nanos)
        self.conf_uh = self.caso.le_exph(self.conf_uh, self.ano_ini, self.nanos)
        self.ree = self.caso.le_ree(self.ree)
        # Cria Sistema
        self.dger = dadosgerais(self.nanos)
        self.submercado, self.intercambio, self.npmc = self.caso.le_sistema(self.submercado, self.intercambio,
                                                                            self.nanos, self.npmc)

        # Calcula produtibilidades acumuladas
        for iusina in self.conf_uh:
            iusina.ProdAcum(self.conf_uh)

        # Calcula Energias Armazenadas
        for ires in self.ree:
            ires.CalcEArmMax(self.conf_uh)
            ires.CalcENA(self.conf_uh)

        for isist in self.submercado:
            if isist.Ficticio == 0:
                isist.CalcEArmMax(self.conf_uh)
                isist.CalcENA(self.conf_uh)

    #############################################################################
    #############################################################################
    # Programacao Dinamica Estocastica Sequencial e Multiprocessing
    #############################################################################
    #############################################################################

    def pde(self, nr_discret, nr_estagios, nr_cenarios):

        t = timer()

        solvers.options['show_progress'] = True
        solvers.options['glpk'] = dict(msg_lev='GLP_MSG_OFF')

        # Define estados a serem visitados em cada estagio
        estados = []
        for i in range(nr_discret):
            estados.append(i / (nr_discret - 1))

        if nr_estagios > 12:
            print('Este exemplo didatico soh funciona para no maxima 12 estagios. Desculpe.')
            return

        # Cria Funcao objetivo, restricoes e inicializa rest. atend. dmenada
        objetivo = 0
        restricoes = []
        atendimentodemanda = 0

        # Cria variaveis de decisao Volume Final, Vertimento e Turbinamento
        vf = variable(len(self.conf_uh), 'vf')
        vv = variable(len(self.conf_uh), 'vv')
        vt = variable(len(self.conf_uh), 'vt')
        for i, iusina in enumerate(self.conf_uh):
            objetivo = objetivo + 0.001 * vv[i]
            # Limite inferior das variaveis hidraulicas
            restricoes.append(vf[i] >= iusina.VolMin)
            restricoes.append(vt[i] >= 0)
            restricoes.append(vv[i] >= 0)
            # Limite superior das variaveis hidraulicas
            restricoes.append(vf[i] <= iusina.VolMax)
            restricoes.append(vt[i] <= iusina.Engolimento)
            restricoes.append(vv[i] <= 100000)
            # Restricao de atendimento a demanda
            prod = float(iusina.Ro65[0][0])
            atendimentodemanda = atendimentodemanda + prod * vt[i]

        # Cria variaveis de decisao Geracao Termica
        gt = variable(len(self.conf_ut), 'gt')
        for i, iusina in enumerate(self.conf_ut):
            objetivo = objetivo + float(iusina.Custo[0]) * gt[i]
            # Restricao de atendimento a demanda
            atendimentodemanda = atendimentodemanda + gt[i]
            # Limite inferior de gt
            restricoes.append(gt[i] >= 0)
            # Limite superior de gt
            restricoes.append(gt[i] <= iusina.Potencia)

        # Cria variaveis de decisao associadas ao deficit
        deficit = variable(1, 'deficit')
        objetivo = objetivo + self.cdef * deficit[0]
        restricoes.append(deficit >= 0)
        restricoes.append(deficit <= 100000)
        atendimentodemanda = atendimentodemanda + deficit[0]

        # Cria Variavel de decisao associada ao custo futuro
        alpha = variable(1, 'alpha')
        objetivo = objetivo + alpha[0]
        restricoes.append(alpha >= 0)
        restricoes.append(alpha <= 10000000)

        restricoes.append(atendimentodemanda == self.mercado)

        # Calcula funcoes objetivos por estado
        for imes in range(nr_estagios, 0, -1):  # Loop de estagio
            print('###############', imes)

            # Cria Função de Custo Futuro Vazia Associada ao Estagio (imes)
            self.cfuturo.append(fcf(imes))

            if len(self.cfuturo) > 1:
                rest_cortes = self.cfuturo[len(self.cfuturo) - 2].get_fcf(vf, alpha)
            else:
                rest_cortes = []

            apontador = len(restricoes)

            # Gera discretizacoes
            discret = product(estados, repeat=len(self.conf_uh))
            for idisc, idiscret in enumerate(discret):  # Loop de discretizacao
                # Inicializa volumes (estados) das usinas correspondentes a iteracao corrente
                VI = matrix(0, (len(self.conf_uh), 1), 'd')
                LAMBDA = matrix(0, (len(self.conf_uh), 1), 'd')
                CUSTO = 0
                for i, iusina in enumerate(self.conf_uh):
                    VI[i] = iusina.VolMin + iusina.VolUtil * idiscret[i]
                    for icen in range(1, nr_cenarios + 1):  # Loop de cenario
                        AFL = matrix(0, (len(self.conf_uh), 1), 'd')
                        for i, iusina in enumerate(self.conf_uh):
                            AFL[i] = float(iusina.Vazoes[icen - 1][imes - 1])
                            restricoes.append(vf[i] == VI[i] + 2.592 * AFL[i] - 2.592 * vt[i] - 2.592 * vv[i])

                        todas = self.reuni(restricoes, rest_cortes)
                        problema = op(objetivo, todas)

                        problema.solve('dense', 'glpk')

                        if problema.status == 'optimal':
                            for i, iusina in enumerate(self.conf_uh):
                                LAMBDA[i] = LAMBDA[i] + todas[apontador + i].multiplier.value
                            CUSTO = CUSTO + problema.objective.value()[0]
                            for i, iusina in enumerate(self.conf_uh):
                                CUSTO = CUSTO - vv[i].value() * 0.001
                        else:
                            print('oops')
                            print(problema.status)
                            problema.tofile('andre.mps')
                            return

                        # Apaga restricoes de Balanco Hidrico
                        ultimo = len(restricoes) - 1
                        for i, iusina in enumerate(self.conf_uh):
                            del restricoes[ultimo]
                            ultimo = ultimo - 1

                    self.cfuturo[len(self.cfuturo) - 1].add_corte(LAMBDA, CUSTO, VI, nr_cenarios)

            # self.plota_corte(imes, nr_discret)
        print(timer() - t)

    def reuni(self, restricoes, rest_cortes):
        tudo = []
        for i in range(0, len(restricoes)):
            tudo.append(restricoes[i])
        for i in range(0, len(rest_cortes)):
            tudo.append(rest_cortes[i])
        return tudo

    def pde_par(self, nr_discret, nr_estagios, nr_cenarios):

        t = timer()

        solvers.options['show_progress'] = False
        solvers.options['glpk'] = dict(msg_lev='GLP_MSG_OFF')

        # Define estados a serem visitados em cada estagio
        estados = []
        for i in range(nr_discret):
            estados.append(i / (nr_discret - 1))

        if nr_estagios > 12:
            print('Este exemplo didatico soh funciona para no maxima 12 estagios. Desculpe.')
            return

        # Cria Funcao objetivo, restricoes e inicializa rest. atend. dmenada
        objetivo = 0
        restricoes = []
        atendimentodemanda = 0

        # Cria variaveis de decisao Volume Final, Vertimento e Turbinamento
        vf = variable(len(self.conf_uh), 'vf')
        vv = variable(len(self.conf_uh), 'vv')
        vt = variable(len(self.conf_uh), 'vt')
        for i, iusina in enumerate(self.conf_uh):
            objetivo = objetivo + 0.001 * vv[i]
            # Limite inferior das variaveis hidraulicas
            restricoes.append(vf[i] >= iusina.VolMin)
            restricoes.append(vt[i] >= 0)
            restricoes.append(vv[i] >= 0)
            # Limite superior das variaveis hidraulicas
            restricoes.append(vf[i] <= iusina.VolMax)
            restricoes.append(vt[i] <= iusina.Engolimento)
            restricoes.append(vv[i] <= 100000)
            # Restricao de atendimento a demanda
            atendimentodemanda = atendimentodemanda + iusina.Ro65 * vt[i]

        # Cria variaveis de decisao Geracao Termica
        gt = variable(len(self.conf_ut), 'gt')
        for i, iusina in enumerate(self.conf_ut):
            objetivo = objetivo + float(iusina.Custo[0]) * gt[i]
            # Restricao de atendimento a demanda
            atendimentodemanda = atendimentodemanda + gt[i]
            # Limite inferior de gt
            restricoes.append(gt[i] >= 0)
            # Limite superior de gt
            restricoes.append(gt[i] <= iusina.Potencia)

        # Cria variaveis de decisao associadas ao deficit
        deficit = variable(1, 'deficit')
        objetivo = objetivo + self.cdef * deficit[0]
        restricoes.append(deficit >= 0)
        restricoes.append(deficit <= 100000)
        atendimentodemanda = atendimentodemanda + deficit[0]

        # Cria Variavel de decisao associada ao custo futuro
        alpha = variable(1, 'alpha')
        objetivo = objetivo + alpha[0]
        restricoes.append(alpha >= 0)
        restricoes.append(alpha <= 10000000)

        restricoes.append(atendimentodemanda == self.mercado)

        # Calcula funcoes objetivos por estado
        for imes in range(nr_estagios, 0, -1):  # Loop de estagio
            print('###############', imes)

            # Cria Função de Custo Futuro Vazia Associada ao Estagio (imes)
            self.cfuturo.append(fcf(imes))

            if len(self.cfuturo) > 1:
                rest_cortes = self.cfuturo[len(self.cfuturo) - 2].get_fcf(vf, alpha)
            else:
                rest_cortes = []

            apontador = len(restricoes)

            # Gera discretizacoes
            discret = product(estados, repeat=len(self.conf_uh))
            C = list(discret)  # Loop de discretizacao

            parametros = {'Sist': self,
                          'nr_cenarios': nr_cenarios,
                          'imes': imes,
                          'restricoes': restricoes,
                          'rest_cortes': rest_cortes,
                          'objetivo': objetivo,
                          'apontador': apontador,
                          'vv': vv,
                          'vt': vt,
                          'vf': vf}

            # self.TaskFcn( parametros, C[0] )
            pool = Pool(processes=4)

            funcao = partial(self.TaskFcn, param=parametros)

            # Resolve com multiprocessing
            res = pool.map(funcao, C)
            # res = map(funcao, C)

            for solucao in res:
                LAMBDA = solucao[2]
                CUSTO = solucao[1]
                VI = solucao[3]
                self.cfuturo[len(self.cfuturo) - 1].add_corte(LAMBDA, CUSTO, VI, nr_cenarios)

            # self.plota_corte(imes, nr_discret)

        print(timer() - t)

    def TaskFcn(self, C, param):

        idiscret = C
        nr_cenarios = param['nr_cenarios']
        sist = param['Sist']
        imes = param['imes']
        restricoes = param['restricoes']
        rest_cortes = param['rest_cortes']
        objetivo = param['objetivo']
        apontador = param['apontador']
        vv = param['vv']
        vt = param['vt']
        vf = param['vf']

        # Inicializa volumes (estados) das usinas correspondentes a iteracao corrente
        VI = matrix(0, (len(sist.conf_uh), 1), 'd')
        LAMBDA = matrix(0, (len(sist.conf_uh), 1), 'd')
        CUSTO = 0
        for i, iusina in enumerate(sist.conf_uh):
            VI[i] = iusina.VolMin + iusina.VolUtil * idiscret[i]
            for icen in range(1, nr_cenarios + 1):  # Loop de cenario
                AFL = matrix(0, (len(sist.conf_uh), 1), 'd')
                for i, iusina in enumerate(sist.conf_uh):
                    AFL[i] = float(iusina.Vazoes[icen - 1][imes - 1])
                    restricoes.append(vf[i] == VI[i] + 2.592 * AFL[i] - 2.592 * vt[i] - 2.592 * vv[i])

                todas = sist.reuni(restricoes, rest_cortes)
                problema = op(objetivo, todas)

                problema.solve('dense', 'glpk')

                if problema.status == 'optimal':
                    for i, iusina in enumerate(sist.conf_uh):
                        LAMBDA[i] = LAMBDA[i] + todas[apontador + i].multiplier.value
                    CUSTO = CUSTO + problema.objective.value()[0]
                    for i, iusina in enumerate(sist.conf_uh):
                        CUSTO = CUSTO - vv[i].value() * 0.001
                else:
                    print('oops')
                    print(problema.status)
                    problema.tofile('andre.mps')
                    return

                # Apaga restricoes de Balanco Hidrico
                ultimo = len(restricoes) - 1
                for i, iusina in enumerate(sist.conf_uh):
                    del restricoes[ultimo]
                    ultimo = ultimo - 1
        return idiscret, CUSTO, LAMBDA, VI

    ####################################################################
    ####################################################################
    # Plotagens Diversas
    ####################################################################
    ####################################################################

    def plota_corte(self, imes, nr_discret):
        for i, ifcf in enumerate(self.cfuturo):
            if ifcf.estagio == imes:
                if len(self.conf_uh) == 1:
                    fig, ax = plt.subplots(figsize=(8, 8))
                    volumes = np.linspace(self.conf_uh[0].VolMin, self.conf_uh[0].VolMax, nr_discret)
                    maior = 0
                    for i in range(0, ifcf.nr_cortes):
                        plt.plot(volumes, (ifcf.coef_vf[i] * volumes + ifcf.termo_i[i]).transpose(), 'b-', lw=3)
                        a = ifcf.coef_vf[i]
                        b = ifcf.termo_i[i]
                        custo = np.array(a * volumes + b)
                        custo.shape = (nr_discret,)
                        if max(custo) > maior:
                            maior = max(custo)
                        plt.fill_between(volumes, 0, custo, facecolor='blue', alpha=0.1)

                    plt.xlabel('Volume Inicial (hm^3)', fontsize=16)
                    titulo = 'Funcao de Custo Futuro do Mes ' + str(imes)
                    plt.title(titulo, fontsize=16)
                    plt.ylabel('Custo (R$)', fontsize=16)
                    plt.xlim(self.conf_uh[0].VolMin, self.conf_uh[0].VolMax)
                    # plt.ylim(0,max(ifcf.termo_i))
                    plt.ylim(0, maior)
                    plt.show()
                    return
                else:
                    print('Somente imprime fcf se existir apenas 1 uh')

    # Plota Usinas Não Existentes e Existentes em Expansao
    def Plota_Expansao_Uh(self):

        # Conta quantas usinas estao
        cont = 0
        nomes = []
        for usina in self.conf_uh:
            if usina.Status == 'EE' or usina.Status == 'NE':
                cont += 1
                nomes.append(usina.Nome)

        motorizada = np.zeros(cont)
        vazia = np.zeros(cont)
        enchendo = np.zeros(cont)
        submotorizada = np.zeros(cont)

        ind = np.arange(cont)
        cont = 0
        for usina in self.conf_uh:
            if usina.Status == 'EE' or usina.Status == 'NE':
                # Meses em que a usina esta motorizada
                motorizada[cont] = self.nanos * 12 - np.count_nonzero(usina.StatusMotoriz - 2)

                # Meses que a usina ainda nao iniciou o enchimento do volume morto
                vazia[cont] = self.nanos * 12 - np.count_nonzero(usina.StatusVolMorto)

                # Meses que a usina encontra-se enchendo o volume morto
                enchendo[cont] = self.nanos * 12 - np.count_nonzero(usina.StatusVolMorto - 1)

                # Meses que a usina encontra-se motorizando
                submotorizada[cont] = self.nanos * 12 - np.count_nonzero(usina.StatusMotoriz - 1)

                cont += 1

        width = 0.35  # the width of the bars: can also be len(x) sequence

        ax = plt.axes()
        p1 = plt.barh(ind, vazia, width, color='w')
        p2 = plt.barh(ind, enchendo, width, color='lime', left=vazia)
        p3 = plt.barh(ind, submotorizada, width, color='sienna', left=vazia + enchendo)
        p4 = plt.barh(ind, motorizada, width, color='black', left=vazia + enchendo + submotorizada)

        plt.ylabel('Usinas', fontsize=16)
        plt.title('Usinas Hidreletricas em Expansao', fontsize=16)
        plt.yticks(ind, nomes, fontsize=12)
        plt.xticks(np.arange(0, self.nanos * 12 + 2, 12))
        # plt.yticks(np.arange(0, 81, 10))
        plt.legend((p1[0], p2[0], p3[0], p4[0]), ('Nao Entrou', 'Enchendo Vol. Morto', 'Submotorizada', 'Motorizada'),
                   fontsize=12)
        plt.xlabel('Meses do Estudo', fontsize=16)
        ax.xaxis.grid()

        plt.show()

    # Plota Mercado de Todos os Submercados
    def PlotaMercado(self):

        f, (ax) = plt.subplots(1, 1)

        nr_lin = len(self.submercado[0].Mercado) - 1
        nr_col = len(self.submercado[0].Mercado[0])
        total = np.zeros((nr_lin, nr_col), 'd')

        for mercado in self.submercado:
            if mercado.Ficticio == 0:
                if mercado.Codigo == 1:
                    cor = 'lime'
                    linha = '--'
                    LineWidth = 2
                elif mercado.Codigo == 2:
                    cor = 'blue'
                    linha = '--'
                    LineWidth = 2
                elif mercado.Codigo == 3:
                    cor = 'chocolate'
                    linha = '--'
                    LineWidth = 2
                elif mercado.Codigo == 4:
                    cor = 'black'
                    linha = '-'
                    LineWidth = 3
                else:
                    cor = 'orange'
                y = np.arange(1, nr_lin * nr_col + 1)
                y = mercado.Mercado[0:nr_lin][0:nr_col]
                ax.plot(np.arange(1, nr_lin * nr_col + 1), (total + y).reshape(nr_lin * nr_col, ), linha, color=cor,
                        lw=LineWidth, label=mercado.Nome)
                ax.fill_between(np.arange(1, nr_lin * nr_col + 1), total.reshape(nr_lin * nr_col, ),
                                (total + y).reshape(nr_lin * nr_col, ), facecolor=cor, alpha=0.1)
                total += y

        titulo = 'Evolucao da Demanda Total do Sistema'
        f.canvas.set_window_title(titulo)

        ax.set_title(titulo, fontsize=16)
        ax.set_xlabel('Meses de Estudo', fontsize=14)
        ax.set_ylabel('Demanda de Energia (MWmes)', fontsize=14)
        ax.legend(fontsize=12)

        plt.show()

    # Plota Energia Armazenada Maxima de Todos REEs
    def PlotaEArmMaxRee(self):

        f, (ax) = plt.subplots(1, 1)
        nr_lin = len(self.ree[0].EArmMax)
        nr_col = len(self.ree[0].EArmMax[0])
        total = np.zeros((nr_lin, nr_col), 'd')

        for ree in self.ree:
            linha = '--'
            LineWidth = 2
            if ree.Codigo == 0:
                cor = 'lime'
            elif ree.Codigo == 1:
                cor = 'blue'
            elif ree.Codigo == 2:
                cor = 'chocolate'
            elif ree.Codigo == 3:
                cor = 'darkgreen'
            elif ree.Codigo == 4:
                cor = 'fuchsia'
            elif ree.Codigo == 5:
                cor = 'gold'
            elif ree.Codigo == 6:
                cor = 'maroon'
            elif ree.Codigo == 7:
                cor = 'orangered'
            elif ree.Codigo == 8:
                cor = 'orchid'
            elif ree.Codigo == 9:
                cor = 'violet'
            elif ree.Codigo == 10:
                cor = 'yellowgreen'
            elif ree.Codigo == 11:
                cor = 'tan'
            else:
                cor = 'black'

            y = np.arange(1, nr_lin * nr_col + 1)
            y = ree.EArmMax[0:nr_lin][0:nr_col]
            ax.plot(np.arange(1, nr_lin * nr_col + 1), (total + y).reshape(nr_lin * nr_col, ), linha, color=cor,
                    lw=LineWidth, label=ree.Nome)
            ax.fill_between(np.arange(1, nr_lin * nr_col + 1), total.reshape(nr_lin * nr_col, ),
                            (total + y).reshape(nr_lin * nr_col, ), facecolor=cor, alpha=0.1)
            total += y

        titulo = 'Evolução da Energia Armazenada Máxima \n Por Reservatorio Equivalente de Energia'
        f.canvas.set_window_title(titulo)

        ax.set_title(titulo, fontsize=16)
        ax.set_xlabel('Meses de Estudo', fontsize=14)
        ax.set_ylabel('Energia Armazenada Maxima (MWmes)', fontsize=14)
        ax.legend(fontsize=12)

        plt.show()

    # Plota Energia Armazenada Maxima de Todos Submercados
    def PlotaEArmMaxSist(self):

        f, (ax) = plt.subplots(1, 1)

        nr_lin = len(self.submercado[0].Mercado) - 1
        nr_col = len(self.submercado[0].Mercado[0])
        total = np.zeros((nr_lin, nr_col), 'd')

        for mercado in self.submercado:
            if mercado.Ficticio == 0:
                linha = '--'
                LineWidth = 2
                if mercado.Codigo == 1:
                    cor = 'lime'
                elif mercado.Codigo == 2:
                    cor = 'blue'
                elif mercado.Codigo == 3:
                    cor = 'chocolate'
                elif mercado.Codigo == 4:
                    cor = 'darkgreen'
                else:
                    cor = 'acqua'

                y = np.arange(1, nr_lin * nr_col + 1)
                y = mercado.EArmMax[0:nr_lin][0:nr_col]
                ax.plot(np.arange(1, nr_lin * nr_col + 1), (total + y).reshape(nr_lin * nr_col, ), linha, color=cor,
                        lw=LineWidth, label=mercado.Nome)
                ax.fill_between(np.arange(1, nr_lin * nr_col + 1), total.reshape(nr_lin * nr_col, ),
                                (total + y).reshape(nr_lin * nr_col, ), facecolor=cor, alpha=0.1)
                total += y

        titulo = 'Evolução da Energia Armazenada Máxima \n Por Submercado'
        f.canvas.set_window_title(titulo)

        ax.set_title(titulo, fontsize=16)
        ax.set_xlabel('Meses de Estudo', fontsize=14)
        ax.set_ylabel('Energia Armazenada Maxima (MWmes)', fontsize=14)
        ax.legend(fontsize=12)

        plt.show()



import numpy as np

class interc(object):

    # Classe contendo informacoes sobre os arquivos de entrada e saida
    De = None
    Para = None
    LimiteMaximo = None
    LimiteMinimo = None
    Flag = None     # Flag 0: Limite de intercambio; 1: Limite de intercambio minimo obrigatorio

    def __init__(self, nanos):
        self.LimiteMaximo = np.zeros((nanos,12), 'd')
        self.LimiteMinimo = np.zeros((nanos,12), 'd')



import numpy as np
from matplotlib import pyplot as plt

class dadosgerais(object):

    # Dados de cadastro das usinas termeletrica (presentes no TERM.DAT)
    NomedoCaso = None
    NAnosEstudo = None
    MesInicioEstudo = None
    MesInicioPreEstudo = None
    AnoInicioEstudo = None
    NAnosPreEstudo = None
    NAnosPosEstudo = None
    NAnosPosSimulacao = None
    NumMaxIter = None
    NumSimForward = None
    NumAberturas = None
    NumSeriesSinteticas = None
    OrdemMaxParP = None
    AnoInicHistorico = None

    def __init__(self, nanos):
        self.NAnosEstudo = nanos








import numpy as np
from mpl_toolkits.mplot3d import axes3d
#import matplotlib
#matplotlib.use('Agg')
from matplotlib import pyplot as plt
from math import sqrt
from random import randint
from pyswarm import pso
from scipy.optimize import minimize
from typing import List

from functools import partial
from scipy.stats import kurtosis

class hidr(object):

    # Dados de cadastro das usinas hidreletricas (presentes no HIDR.DAT)
    Codigo = None           # Codigo da UHE
    Nome = None             # Nome da UHE
    Posto = None            # Numero do Posto
    Bdh = None              # Desvio - Nao sei qual e esta informacao ??????
    Sist = None             # Submercado
    Empr = None             # Codigo da empresa
    Jusante = None          # Codigo de Jusante
    Desvio = None           # Desvio - Nao sei qual e esta informacao ??????
    VolMin = None           # Volume Minimo
    VolMax = None           # Volume Maximo
    VolVert = None          # Volume Vertimento
    VolMinDesv = None       # Volume Minimo para Desvio
    CotaMin = None          # Cota Minima
    CotaMax = None          # Cota Maxima
    PolCotaVol = None       # Polinomio Cota-Volume
    PolCotaArea = None      # Polinomio Cata-Area
    CoefEvap = None         # Coeficientes de Evaporacao
    NumConjMaq = None       # Numero de Conjuntos de Maquinas
    MaqporConj = None       # Numero de Maquinas por Conjunto
    PEfporConj = None       # POtencia Efetiva por Maquina do Conjunto

    CF_HBQT = None          # Nao sei qual e esta informacao ??????
    CF_HBQG = None          # Nao sei qual e esta informacao ??????
    CF_HBPT = None          # Nao sei qual e esta informacao ??????

    AltEfetConj = None      # Altura de Queda Efetiva do Conjunto
    VazEfetConj = None      # Vazao Efetiva do Conjunto
    ProdEsp = None          # Produtibilidade Especifica
    PerdaHid = None         # Perda Hidraulica
    NumPolVNJ = None        # Numero de Polinomios Vazao Nivel Jusante

    PolVazNivJus = None     # Polinomios Vazao Nivel Jusante

    CotaRefNivelJus = None  # Cota Referencia Nivel de Jusante
    CFMed = None            # Cota Media do Canal de Fuga
    InfCanalFuga = None     # Informacao Canal de Fuga - Nao sei qual e esta informacao ??????
    FatorCargaMax = None    # Fator de Caga Maximo - Nao sei qual e esta informacao ?????????
    FatorCargaMin = None    # Fator de Caga Minimo - Nao sei qual e esta informacao ?????????
    VazMin = None           # Vazao Minima Obrigatoria
    UnidBase = None         # Numero de Unidades de Base
    TipoTurb = None         # Tipo de Turbina Hidraulica
    Repres_Conj = None      # Representacao Conjunto de Maquina - Nao sei qual e esta informacao ?????
    TEIFH = None            # Taxa Equivalente de Indisponibilidade Forcada Hidraulica
    IP = None               # Indisponibilidade Programada
    TipoPerda = None        # Tipo Perda Hidraulica
    Data = None             # Nao sei qual e esta informacao ??????
    Observ = None           # Observacao
    VolRef = None           # Volume de Referencia
    TipoReg = None          # Tipo de Regulacao

    # Dados Adicionais Especificados no arquivo de configuracao hidraulica (CONFHD)
    Ree = None
    Status = None
    VolIni = None
    Modif = None
    AnoI = None
    AnoF = None

    # Dados Adicinais Calculados para as Usinas pertecentes a configuracao hidraulica (CONFHD)
    VolUtil = None
    VazEfet = None
    PotEfet = None
    Ro65 = None             # PDTMED (NEWAVE) - PROD. ASSOCIADA A ALTURA CORRESPONDENTE A 65% DO V.U.
    Ro50 = None
    RoMax = None            # PDTMAX (NEWAVE) - PROD. ASSOCIADA A ALTURA MAXIMA
    RoMin = None            # PDTMIN (NEWAVE) - PROD. ASSOCIADA A ALTURA MINIMA
    RoEquiv = None          # PRODT (NEWAVE) - PROD. EQUIVALENTE ( DO VOL. MINIMO AO VOL. MAXIMO )
    RoEquiv65 = None        # PRODTM (NEWAVE) - PROD. EQUIVALENTE ( DO VOL. MINIMO A 65% DO V.U. )
    Engolimento = None
    RoAcum = None           # PDTARM (NEWAVE) - PROD. ACUM. PARA CALCULO DA ENERGIA ARMAZENADA
    RoAcum65 = None         # PDAMED (NEWAVE) - PROD. ACUM. PARA CALCULO DA ENERGIA ARMAZENADA CORRESPONDENTE A 65% DO V.U.
    RoAcumMax = None        # PDCMAX e PDVMAX (NEWAVE) - PROD. ACUM.
    RoAcumMed = None        # PDTCON, PDCMED e PDVMED (NEWAVE) - PROD. ACUM.
    RoAcumMin = None        # PDCMIN e PDVMIN (NEWAVE) - PROD. ACUM.

    RoAcum_A_Ree = None
    RoAcum_B_Ree = None
    RoAcum_C_Ree = None
    RoAcum_A_Sist = None
    RoAcum_B_Sist = None
    RoAcum_C_Sist = None

    RoAcumEntreResRee = None
    RoAcumEntreResSist = None

    # Vazoes Naturais, Incrementais e Par(p)
    Vazoes = None       # Historico de Vazoes naturais (imes, ilag)
    FAC = None          # Funcao de Autocorrelacao (imes, ilag)
    FACP = None         # Funcao de Autocorrelacao Parcial (imes, ilag)
    CoefParp = None     # Coeficientes do Modelo par(p) (imes,ilag)
    CoefIndParp  = None     # Coeficientes independentes do Modelo par(p) (imes) - Aditivo = 0 - Multiplicativo > 0
    Ordem = None        # Ordem do modelo par(p) para todos os meses (mes)

    # Parametros da usina Dependentes do Tempo - Especificados (MODIF.DAT)
    VolMinT = None     # Volume Mínimo Operativo (pode variar mes a mes)
    VolMaxT = None     # Volume Maximo Operativo (pode variar mes a mes)
    VolMinP = None     # Volume Mínimo com adocao de penalidade (pode variar mes a mes)
    VazMinT = None     # Vazao Minima pode variar mes a mes
    CFugaT  = None     # Cota do Canal de Fuga (pode varia mes a mes)

    # Parametros relativos a expansao hidrica que variam no tempo para usinas 'EE' e 'NE' (EXPH)
    StatusVolMorto = None       # Status do Volume Morto - 0: Nao Comecou Encher - 1: Enchendo - 2: Cheio
    VolMortoTempo = None        # Evolucao do Volume Minimo da Usina
    StatusMotoriz  = None       # Status da Motorizacao  - 0: Nao Comecou Motorizar - 1: Motorizando - 3: Motorizada
    UnidadesTempo = None        # Numero de Unidades em cada mes
    EngolTempo = None           # Evolucao do Engolimento Maximo da Usina
    PotenciaTempo = None        # Evolucao da Potencia Instalada da Usina

    ##########################################################################################################
    # Graficos Diversos
    ##########################################################################################################

    # Plota Polinomio Cota-Volume
    def PlotaPCV(self):
        if self.VolMin == 0:
            return

        if (self.VolMin == self.VolMax):
            volumes = np.linspace(self.VolMin - 1,self.VolMax + 1, 100)
        else:
            volumes = np.linspace(self.VolMin,self.VolMax,100)
        a = self.PolCotaVol[0]
        b = self.PolCotaVol[1]
        c = self.PolCotaVol[2]
        d = self.PolCotaVol[3]
        e = self.PolCotaVol[4]
        cota = a + b*volumes + c*volumes**2 + d*volumes**3 + e*volumes**4
        cota.shape = volumes.shape
        plt.plot(volumes, cota, 'b-', lw=3)

        plt.xlabel('Volume do Reservatorio (hm^3)', fontsize=16)
        titulo = 'Polinomio Cota-Volume da Usina ' + self.Nome
        plt.title(titulo, fontsize=16)
        plt.ylabel('Cota em Metros', fontsize=16)
        plt.xlim(volumes[0], volumes[99])
        if ( cota[0] == cota[99]):
            plt.ylim(cota[0]-1, cota[99]+1)
        else:
            plt.ylim(cota[0], cota[99])
        plt.show()

    # Plota Polinomio Cota-Area
    def PlotaPCA(self):
        if self.VolMin == 0:
            return

        if (self.CotaMax == self.CotaMin):
            cotas = np.linspace(self.CotaMin - 1,self.CotaMax + 1, 100)
        else:
            cotas = np.linspace(self.CotaMin,self.CotaMax,100)
        a = self.PolCotaArea[0]
        b = self.PolCotaArea[1]
        c = self.PolCotaArea[2]
        d = self.PolCotaArea[3]
        e = self.PolCotaArea[4]
        areas = a + b*cotas + c*cotas**2 + d*cotas**3 + e*cotas**4
        areas.shape = cotas.shape
        plt.plot(cotas, areas, 'b-', lw=3)

        plt.xlabel('Cota do Reservatorio (em metros)', fontsize=16)
        titulo = 'Polinomio Cota-Area da Usina ' + self.Nome
        plt.title(titulo, fontsize=16)
        plt.ylabel('Area Superficia em km^2', fontsize=16)
        plt.xlim(cotas[0], cotas[99])
        if ( areas[0] == areas[99]):
            plt.ylim(areas[0]-1, areas[99]+1)
        else:
            plt.ylim(areas[0], areas[99])
        plt.show()

    # Plota Curva Colina
    def PlotaColina(self):
        if self.VolMin == 0:
            return

        if (self.VolMin == self.VolMax):
            volumes = np.linspace(self.VolMin - 1,self.VolMax + 1, 100)
        else:
            volumes = np.linspace(self.VolMin,self.VolMax,100)

        a = self.PolCotaVol[0]
        b = self.PolCotaVol[1]
        c = self.PolCotaVol[2]
        d = self.PolCotaVol[3]
        e = self.PolCotaVol[4]

        cotamont = a + b*volumes + c*volumes**2 + d*volumes**3 + e*volumes**4
        cotamont.shape = volumes.shape

        qdef = np.linspace(self.VazMin, 5*self.Engolimento, 100)

        a = self.PolVazNivJus[0][0]
        b = self.PolVazNivJus[0][1]
        c = self.PolVazNivJus[0][2]
        d = self.PolVazNivJus[0][3]
        e = self.PolVazNivJus[0][4]

        cotajus = a + b*qdef + c*qdef**2 + d*qdef**3 + e*qdef**4
        cotajus.shape = qdef.shape

        xGrid, yGrid = np.meshgrid(cotamont, cotajus)

        z = self.ProdEsp * ( xGrid - yGrid )

        fig = plt.figure()
        ax = fig.gca(projection='3d')

        surf = ax.plot_surface(qdef, volumes,z, rcount=100, ccount = 100, cmap=plt.cm.coolwarm,
                       linewidth=0, antialiased=False)

        plt.xlabel('Vazão Defluente em m^3/s', fontsize=12)
        titulo = 'Produtibilidade da Usina ' + self.Nome
        plt.title(titulo, fontsize=16)
        plt.ylabel('Volume Armazenado em hm^3', fontsize=12)
        fig.colorbar(surf, shrink=0.5, aspect=5)

        plt.show()


    def PlotaProdutibs(self, iano, imes):
        x_axis = np.arange(1,6)
        y_axis = [ self.RoEquiv[iano][imes], self.RoMin[iano][imes], self.Ro50[iano][imes], self.Ro65[iano][imes], self.RoMax[iano][imes] ]
        fig, ax = plt.subplots()
        a, b, c, d, e = plt.bar(x_axis, y_axis)
        a.set_facecolor('r')
        b.set_facecolor('g')
        c.set_facecolor('b')
        d.set_facecolor('y')
        e.set_facecolor('m')
        ax.set_xticks(x_axis)
        ax.set_xticklabels(['Equiv', 'Min', '50%', '65%', 'Max'])
        titulo = 'Produtibilidades da Usina ' + self.Nome
        plt.title(titulo, fontsize=16)
        plt.xlabel('Tipo de Produtibilidade', fontsize=16)
        plt.ylabel('Produtibilidade', fontsize=16)
        plt.show()

    def PlotaVazoes(self):
        x_axis = np.arange(1,13)
        plt.plot(x_axis,self.Vazoes.transpose(),'c-')
        media = np.mean(self.Vazoes, axis=0)
        plt.plot(x_axis,media,'r-',lw=3)
        desvio = np.nanstd(self.Vazoes, axis=0)
        plt.plot(x_axis,media+desvio,'r-.',lw=2)
        plt.plot(x_axis,media-desvio,'r-.',lw=2)
        ultimo = len(self.Vazoes)-1
        plt.plot(x_axis,self.Vazoes[:][ultimo],'b-')
        titulo = 'Historico de Vazoes da Usina ' + self.Nome
        plt.title(titulo, fontsize=16)
        plt.xlabel('Mes do Ano', fontsize=16)
        plt.ylabel('Vazao', fontsize=16)
        plt.show()

    def PlotaVolume(self):
        nanos = len(self.VolMinT)

        fig = plt.figure()
        ax = plt.subplot(111)


        x_axis = np.arange(1,nanos*12+1)
        ax.plot(x_axis,self.VolMinT.reshape(nanos*12),'g-.',lw=2, label = 'Vol.Min.Operat.')
        ax.plot(x_axis,self.VolMaxT.reshape(nanos*12),'g-.',lw=2, label = 'Vol.Max.Operat.')
        ax.plot(x_axis,self.VolMax*np.ones(nanos*12),'b-',lw=3,   label = 'Vol.Minimo Real')
        ax.plot(x_axis,self.VolMin*np.ones(nanos*12),'b-',lw=3,   label = 'Vol.Maximo Real')
        ax.plot(x_axis,self.VolMinP.reshape(nanos*12),'b-.',lw=2, label = 'Vol.Min.com Pen.')

        plt.fill_between(x_axis,self.VolMinT.reshape(nanos*12), self.VolMaxT.reshape(nanos*12), facecolor='g', alpha=0.1)

        titulo = 'Evolucao dos Volumes da Usina \n' + self.Nome
        plt.title(titulo, fontsize=16)
        plt.xlabel('Mes de Estudo', fontsize=16)
        plt.ylabel('Volume em hm^3', fontsize=16)

        box = ax.get_position()

        ax.set_position([ box.x0, box.y0, box.width*0.7, box.height] )

        ax.legend(loc='center left', shadow=True, fontsize=12, bbox_to_anchor=(1, 0.5))

        plt.show()

    def PlotaVazMin(self):
        nanos = len(self.VazMinT)

        fig = plt.figure()
        ax = plt.subplot(111)

        x_axis = np.arange(1,nanos*12+1)
        ax.plot(x_axis,self.VazMinT.reshape(nanos*12),'g-.',lw=2, label='Vaz.Min.Operat.')
        ax.plot(x_axis,self.VazMin*np.ones(nanos*12),'b-',lw=3,   label='Vaz.Min.Cadastro')

        titulo = 'Evolucao da Vazao Minima da Usina \n' + self.Nome
        plt.title(titulo, fontsize=16)
        plt.xlabel('Mes de Estudo', fontsize=16)
        plt.ylabel('Vazao Minima em m^3', fontsize=16)

        box = ax.get_position()

        ax.set_position([ box.x0, box.y0, box.width*0.7, box.height] )

        ax.legend(loc='center left', shadow=True, fontsize=12, bbox_to_anchor=(1, 0.5))

        plt.show()

    def PlotaVolMorto(self):

        if self.Status == 'EX':
            print('Grafico de Volume Morto nao impresso, pois ', self.Nome, 'e uma usina existente')
            return

        nanos = len(self.VolMortoTempo)

        nmeses = np.count_nonzero(self.VolMortoTempo)

        legenda = str(nmeses) + ' Meses'

        ax = plt.subplot(111)

        x_axis = np.arange(1,nanos*12+1)
        p1 = ax.plot(x_axis,self.VolMortoTempo.reshape(nanos*12),'g-.',lw=2, label = legenda )

        titulo = 'Enchimento do Volume Morto da Usina \n' + self.Nome
        plt.title(titulo, fontsize=16)
        plt.xlabel('Mes de Estudo', fontsize=16)
        plt.ylabel('Volume Morto em hm^3', fontsize=16)

        plt.legend(fontsize=12)

        np.count_nonzero(self.VolMortoTempo)

        plt.show()

    def PlotaPotencia(self):

        nanos = len(self.PotenciaTempo)

        ax = plt.subplot(111)

        x_axis = np.arange(1, nanos * 12 + 1)
        p1 = ax.plot(x_axis, self.PotenciaTempo.reshape(nanos * 12), 'g-.', lw=2)

        titulo = 'Evolucao da Potencia Efetiva da Usina \n' + self.Nome
        plt.title(titulo, fontsize=16)
        plt.xlabel('Mes de Estudo', fontsize=16)
        plt.ylabel('Potencia Efetiva em MW', fontsize=16)

        plt.show()

    def PlotaParp(self, mes):

        ordmax = len(self.CoefParp[0])
        nanos = len(self.Vazoes) - 1

        if mes == 0:
            str_mes = 'January'
        elif mes == 1:
            str_mes = 'Fevereiro'
        elif mes == 2:
            str_mes = 'Marco'
        elif mes == 3:
            str_mes = 'Abril'
        elif mes == 4:
            str_mes = 'Maio'
        elif mes == 5:
            str_mes = 'Junho'
        elif mes == 6:
            str_mes = 'Julho'
        elif mes == 7:
            str_mes = 'Agosto'
        elif mes == 8:
            str_mes = 'Setembro'
        elif mes == 9:
            str_mes = 'Outubro'
        elif mes == 10:
            str_mes = 'Novembro'
        else:
            str_mes = 'Dezembro'

        IC = 1.96/sqrt(nanos-1)

        cores = []
        limitesup = []
        limiteinf = []
        for elemento in self.FACP[mes][1:ordmax+1]:
            limitesup.append(IC)
            limiteinf.append(-IC)
            if elemento > IC or elemento < -IC:
                cores.append('r')
            else:
                cores.append('b')

        f, (ax1, ax2) = plt.subplots(1, 2, sharey=True)
        barWidth = 0.40


        titulo = 'FAC e FACP of ' + str_mes + ' - UHE ' + self.Nome
        f.canvas.set_window_title(titulo)

        ax1.bar(np.arange(1,ordmax+1), self.FAC[mes][1:ordmax+1], barWidth, align='center')
        ax2.bar(np.arange(1,ordmax+1), self.FACP[mes][1:ordmax+1], barWidth, align='center', color = cores)
        ax2.plot(np.arange(1,ordmax+1), limitesup, 'm--', lw=1)
        ax2.plot(np.arange(1,ordmax+1), limiteinf, 'm--', lw=1)

        ax1.set_xticks(np.arange(1,ordmax+1))
        ax2.set_xticks(np.arange(1,ordmax+1))
        tituloFAC =  'FAC - Month: ' + str_mes + '\n of UHE ' + self.Nome
        tituloFACP = 'FACP - Month ' + str_mes +  '\n of UHE ' + self.Nome
        ax1.set_title(tituloFAC,  fontsize = 13)
        ax2.set_title(tituloFACP, fontsize =13)
        #ax1.xlabel('Lag')
        #ax2.xlabel('Lag')
        #ax1.ylabel('Autocorrelacao e Autocorrelacao Parcial')

        plt.show()


    ##########################################################################################################
    # Calcula Parametros das Usinas
    ##########################################################################################################

    def CalcVolUtil(self):     # Calcula Volume Util da Usina
        if self.TipoReg == b'M':
            self.VolUtil = self.VolMax - self.VolMin
        else:
            self.VolUtil = float(0)
            self.VolMin = self.VolMax

    def CalcPotEfetiva(self):     # Calcula Potencia Efetiva da Usina
        a = np.array(self.MaqporConj)
        b = np.array(self.PEfporConj)
        self.PotEfet = np.vdot(a, b)

    def CalcVazEfetiva(self):      # Calcula Vazao Efetiva da Usina
        a = np.array(self.MaqporConj)
        b = np.array(self.VazEfetConj)
        self.VazEfet = np.vdot(a, b)

    def CalcEngolMaximo(self):    # Estima Engolimento Maximo da Usina

        def CalcEngol(self, ql):
            engol = 0.
            for i in range(5):   # Varre Conjuntos de Maquinas
                if self.MaqporConj[i] > 0:
                    if ql < self.AltEfetConj[i]:
                        if self.TipoTurb == 1 or self.TipoTurb == 3:
                            alpha = 0.5
                        else:
                            alpha = 0.2
                    else:
                        alpha = -1
                    if self.AltEfetConj[i] != 0:
                        engol = engol + self.MaqporConj[i]*self.VazEfetConj[i]*((ql/self.AltEfetConj[i])**alpha)
            return engol

        a = self.PolCotaVol[0]
        b = self.PolCotaVol[1]
        c = self.PolCotaVol[2]
        d = self.PolCotaVol[3]
        e = self.PolCotaVol[4]

        # Calcula Engolimento a 65% do Volume Util
        volume = self.VolMin + 0.65*self.VolUtil
        cota = a + b*volume + c*volume**2 + d*volume**3 + e*volume**4
        queda65 = cota - self.CFMed
        engol65 = CalcEngol(self, queda65)

        # Calcula Engolimento a 50% do Volume Util
        volume = self.VolMin + 0.50*self.VolUtil
        cota = a + b*volume + c*volume**2 + d*volume**3 + e*volume**4
        queda50 = cota - self.CFMed
        engol50 = CalcEngol(self, queda50)

        # Calcula Engolimento Associada ao Volume Maximo
        volume = self.VolMax
        cota = a + b*volume + c*volume**2 + d*volume**3 + e*volume**4
        quedaMax = cota - self.CFMed
        engolMax = CalcEngol(self, quedaMax)

        # Calcula Engolimento Associada ao Volume Minimo
        volume = self.VolMin
        cota = a + b*volume + c*volume**2 + d*volume**3 + e*volume**4
        quedaMin = cota - self.CFMed
        engolMin = CalcEngol(self, quedaMin)

        # Calcula Engolimento Associado a Altura Equivalente
        if ( self.VolUtil > 0):
            cota = 0
            for i in range(5):
                cota = cota + self.PolCotaVol[i] * (self.VolMax**(i+1)) / (i+1)
                cota = cota - self.PolCotaVol[i] * (self.VolMin**(i+1)) / (i+1)
            cota = cota / self.VolUtil
        quedaEquiv = cota - self.CFMed
        engolEquiv = CalcEngol(self, quedaEquiv)

        self.Engolimento = (engol50+engol65+engolEquiv+engolMax+engolMin)/5

        return

    def CalcProdutibs(self, nanos):       # Calcula Produtibilidades Associadas aa diversos volumes

        self.Ro65       = np.zeros( (nanos,12), 'd' )
        self.Ro50       = np.zeros( (nanos,12), 'd' )
        self.RoEquiv    = np.zeros( (nanos,12), 'd' )
        self.RoEquiv65  = np.zeros( (nanos,12), 'd' )
        self.RoMin      = np.zeros( (nanos,12), 'd' )
        self.RoMax      = np.zeros( (nanos,12), 'd' )

        a = self.PolCotaVol[0]
        b = self.PolCotaVol[1]
        c = self.PolCotaVol[2]
        d = self.PolCotaVol[3]
        e = self.PolCotaVol[4]

        # Calcula Produtibilidade Associada a 65% do Volume Util
        volume = self.VolMin + 0.65*self.VolUtil
        cota = a + b*volume + c*volume**2 + d*volume**3 + e*volume**4
        for iano in range(nanos):
            for imes in range(12):
                cfuga = self.CFugaT[iano][imes]
                if self.TipoPerda == 2:
                    self.Ro65[iano][imes] = self.ProdEsp * (cota - cfuga - self.PerdaHid)
                else:
                    self.Ro65[iano][imes] = self.ProdEsp * (cota - cfuga)*(1. - self.PerdaHid/100)

        # Calcula Produtibilidade Associada a 50% do Volume Util
        volume = self.VolMin + 0.50*self.VolUtil
        cota = a + b*volume + c*volume**2 + d*volume**3 + e*volume**4
        for iano in range(nanos):
            for imes in range(12):
                cfuga = self.CFugaT[iano][imes]
                if self.TipoPerda == 2:
                    self.Ro50[iano][imes] = self.ProdEsp * (cota - cfuga - self.PerdaHid)
                else:
                    self.Ro50[iano][imes] = self.ProdEsp * (cota - cfuga)*(1. - self.PerdaHid/100)

        # Calcula Produtibilidade Associada ao Volume Maximo
        volume = self.VolMax
        cota = a + b*volume + c*volume**2 + d*volume**3 + e*volume**4
        for iano in range(nanos):
            for imes in range(12):
                cfuga = self.CFugaT[iano][imes]
                if self.TipoPerda == 2:
                    self.RoMax[iano][imes] = self.ProdEsp * (cota - cfuga - self.PerdaHid)
                else:
                    self.RoMax[iano][imes] = self.ProdEsp * (cota - cfuga)*(1. - self.PerdaHid/100)

        # Calcula Produtibilidade Associada ao Volume Minimo
        volume = self.VolMin
        cota = a + b*volume + c*volume**2 + d*volume**3 + e*volume**4
        for iano in range(nanos):
            for imes in range(12):
                cfuga = self.CFugaT[iano][imes]
                if self.TipoPerda == 2:
                    self.RoMin[iano][imes] = self.ProdEsp * (cota - cfuga - self.PerdaHid)
                else:
                    self.RoMin[iano][imes] = self.ProdEsp * (cota - cfuga)*(1. - self.PerdaHid/100)

        # Calcula Produtibilidade Equivalente
        if ( self.VolUtil > 0):
            cota = 0
            cota65 = 0
            Vol65 = self.VolMin + 0.65*self.VolUtil
            for i in range(5):
                cota = cota + self.PolCotaVol[i] * (self.VolMax**(i+1)) / (i+1)
                cota = cota - self.PolCotaVol[i] * (self.VolMin**(i+1)) / (i+1)
                cota65 = cota65 + self.PolCotaVol[i] * (Vol65**(i+1)) / (i+1)
                cota65 = cota65 - self.PolCotaVol[i] * (self.VolMin**(i+1)) / (i+1)
            cota = cota / self.VolUtil
            cota65 = cota65 / (Vol65 - self.VolMin)
        else:
            cota65 = cota
        for iano in range(nanos):
            for imes in range(12):
                cfuga = self.CFugaT[iano][imes]
                if self.TipoPerda == 2:
                    self.RoEquiv[iano][imes]   = self.ProdEsp * (cota   - cfuga - self.PerdaHid)
                    self.RoEquiv65[iano][imes] = self.ProdEsp * (cota65 - cfuga - self.PerdaHid)
                else:
                    self.RoEquiv[iano][imes]   = self.ProdEsp * (cota   - cfuga)*(1. - self.PerdaHid/100)
                    self.RoEquiv65[iano][imes] = self.ProdEsp * (cota65 - cfuga)*(1. - self.PerdaHid/100)
        return

    # Calcula Vazao Incremental
    def QInc(self, usinas, iano, imes):

        nanos_hist = len(self.Vazoes)

        def Montante(usinas, usina, iano, imes):
            for iusi in usinas:
                if iusi.Jusante == usina.Codigo:
                    if iusi.StatusVolMorto[iano][imes] == 2:
                        yield iusi
                    else:
                        yield from Montante(usinas, iusi, iano, imes)

        if self.StatusVolMorto[iano][imes] != 2:
            print ('Erro: Tentativa de calculo de Incremental para usina (', self.Nome, ') fora de operacao no mes ', imes, ' e ano ', iano)
            return 0
        else:
            Incremental = self.Vazoes[0:nanos_hist,imes]
            for iusina in Montante(usinas, self, iano, imes):
                Incremental = Incremental - iusina.Vazoes[0:nanos_hist,imes]

        if np.min(Incremental) < 0:
            contador = 0
            for i in range(nanos_hist):
                if Incremental[i] < 0:
                    Incremental[i] = 0
                    contador = contador + 1
            return Incremental
        else:
            return Incremental

    # Calcula vazao incremental entre a usina e todos os reservatorios a montante
    def QIncEntreRes(self, usinas, ianoconf, imesconf):

        nanos_hist = len(self.Vazoes)

        def Montante(usinas, usina, iano, imes):
            for iusi in usinas:
                if iusi.Jusante == usina.Codigo:
                    if iusi.StatusVolMorto[iano][imes] == 2:
                        if iusi.VolUtil > 0:
                            yield iusi
                        else:
                            yield from Montante(usinas, iusi, iano, imes)
                    else:
                        yield from Montante(usinas, iusi, iano, imes)

        if self.StatusVolMorto[ianoconf][imesconf] != 2:
            print ('Erro: Tentativa de calculo de Incremental para usina (', self.Nome, ') fora de operacao no mes ', imesconf, ' e ano ', ianoconf)
            return 0
        else:
            Incremental = np.zeros( (nanos_hist,1) , 'd')
            Incremental = self.Vazoes[0:nanos_hist,imesconf]
            for iusina in Montante(usinas, self, ianoconf, imesconf):
                Incremental = Incremental - iusina.Vazoes[0:nanos_hist,imesconf]

        if np.min(Incremental) < 0:
            contador = 0
            for i in range(nanos_hist):
                if Incremental[i] < 0:
                    #Incremental[i] = 0
                    contador = contador + 1
            #print ('Vazao Incremental da Usina ', self.Nome, 'menor que zero no mes ', imesconf, ' e ano ', ianoconf, 'Quantidade:', contador )
            return Incremental
        else:
            return Incremental

    def ProdAcum(self, usinas):

        def Cascata(usinas,iano,imes):
            current = self
            if current.StatusVolMorto[iano][imes] == 2:
                yield current
            while current.Jusante != 0:
                for iusi in usinas:
                    if iusi.Codigo == current.Jusante:
                        if iusi.StatusVolMorto[iano][imes] == 2:
                            yield iusi
                        current = iusi
                        break

        nanos = len(self.StatusVolMorto)

        self.RoAcum_A_Ree = np.zeros( (nanos,12), 'd')
        self.RoAcum_B_Ree = np.zeros( (nanos,12), 'd')
        self.RoAcum_C_Ree = np.zeros( (nanos,12), 'd')

        self.RoAcum_A_Sist = np.zeros( (nanos,12), 'd')
        self.RoAcum_B_Sist = np.zeros( (nanos,12), 'd')
        self.RoAcum_C_Sist = np.zeros( (nanos,12), 'd')

        self.RoAcum    = np.zeros( (nanos,12), 'd')
        self.RoAcum65  = np.zeros( (nanos,12), 'd' )
        self.RoAcumMax = np.zeros( (nanos,12), 'd' )
        self.RoAcumMed = np.zeros( (nanos,12), 'd' )
        self.RoAcumMin = np.zeros( (nanos,12), 'd' )

        for iano in range(nanos):
            for imes in range(12):
                trocouRee = 0
                trocouSist = 0
                FioRee = True
                FioSist = True

                for iusina in Cascata(usinas, iano, imes):
                    produtib    = iusina.RoEquiv[iano][imes]
                    produtib65  = iusina.RoEquiv65[iano][imes]
                    produtibMax = iusina.RoMax[iano][imes]
                    produtibMed = iusina.Ro65[iano][imes]
                    produtibMin = iusina.RoMin[iano][imes]
                    if iusina.StatusMotoriz[iano][imes] == 2:
                        self.RoAcum[iano][imes]    = self.RoAcum[iano][imes]    + produtib
                        self.RoAcum65[iano][imes]  = self.RoAcum65[iano][imes]  + produtib65
                        self.RoAcumMax[iano][imes] = self.RoAcumMax[iano][imes] + produtibMax
                        self.RoAcumMed[iano][imes] = self.RoAcumMed[iano][imes] + produtibMed
                        self.RoAcumMin[iano][imes] = self.RoAcumMin[iano][imes] + produtibMin
                    if iusina.Sist != self.Sist:
                        trocouSist = trocouSist+ 1
                    if iusina.Ree != self.Ree:
                        trocouRee = trocouRee + 1

                    if trocouRee == 0:
                        if iusina.StatusMotoriz[iano][imes] == 2:
                            self.RoAcum_A_Ree[iano][imes] = self.RoAcum_A_Ree[iano][imes] + produtib
                    else:
                        if iusina.VolUtil > 0:
                            FioRee = False
                        if FioRee:
                            if iusina.StatusMotoriz[iano][imes] == 2:
                                self.RoAcum_B_Ree[iano][imes] = self.RoAcum_B_Ree[iano][imes] + produtib
                        else:
                            if iusina.StatusMotoriz[iano][imes] == 2:
                                self.RoAcum_C_Ree[iano][imes] = self.RoAcum_C_Ree[iano][imes] + produtib

                    if trocouSist == 0:
                        if iusina.StatusMotoriz[iano][imes] == 2:
                            self.RoAcum_A_Sist[iano][imes] = self.RoAcum_A_Sist[iano][imes] + produtib
                    else:
                        if iusina.VolUtil > 0:
                            FioSist = False
                        if FioSist:
                            if iusina.StatusMotoriz[iano][imes] == 2:
                                self.RoAcum_B_Sist[iano][imes] = self.RoAcum_B_Sist[iano][imes] + produtib
                        else:
                            if iusina.StatusMotoriz[iano][imes] == 2:
                                self.RoAcum_C_Sist[iano][imes] = self.RoAcum_C_Sist[iano][imes] + produtib

    def ProdAcumEntreResRee(self, iano, imes, usinas):
         if self.Jusante == 0:
             return 0
         for iusina in usinas:
             if iusina.Codigo == self.Jusante:
                 if iusina.VolUtil != 0:
                     return 0.
                 elif self.Ree != iusina.Ree:
                     return 0.
                 elif iusina.StatusMotoriz[iano][imes] == 2:
                     return iusina.RoEquiv[iano][imes] + iusina.ProdAcumEntreResRee(iano, imes, usinas)
                 else:
                     return iusina.ProdAcumEntreResRee(iano, imes, usinas)
                 break
    #
    # def ProdAcumEntreResSist(self, iano, imes, usinas):
    #     if self.Jusante == 0:
    #         return 0
    #     for iusina in usinas:
    #         if iusina.Codigo == self.Jusante:
    #             if iusina.VolUtil != 0:
    #                 return 0.
    #             elif self.Sist != iusina.Sist:
    #                 return 0.
    #             elif iusina.StatusMotoriz[iano][imes] == 2:
    #                 return iusina.RoEquiv + iusina.ProdAcumEntreResSist(iano, imes, usinas)
    #             else:
    #                 return iusina.ProdAcumEntreResSist(iano, imes, usinas)
    #             break



    #########################################################
    # Calcula Modelo PAR(p)
    #########################################################

    def parp(self, ord_max):

        nanos = len(self.Vazoes) - 1  # A serie historica do ultimo ano geralmente nao vem completa (despreze-a)

        media = np.mean(self.Vazoes[1:nanos], 0)    # A primeira serie historica eh utilizada como tendencia (despreze-a)
        desvio = np.std(self.Vazoes[1:nanos], 0)    # A primeira serie historica eh utilizada como tendencia (despreze-a)

        # Calcula vazao normalizada (nao precisa)
        #vaznorm = np.zeros((nanos,12),'d')
        #for iano in range(nanos):
        #    for imes in range(12):
        #        vaznorm[iano][imes] = (self.Vazoes[iano][imes] - media[imes])/desvio[imes]

        # Calcula funcao de auto-correlacao (uma para cada mes)
        self.FAC = np.zeros( (12, ord_max+1), 'd')
        for ilag in range(ord_max+1):
            for imes in range(12):
                for iano in np.arange(1,nanos):
                     ano_ant = iano
                     mes_ant = imes - ilag
                     if mes_ant < 0:
                         ano_ant -= 1
                         mes_ant += 12
                     self.FAC[imes][ilag] += (self.Vazoes[iano][imes] - media[imes])* (self.Vazoes[ano_ant][mes_ant] - media[mes_ant])
                self.FAC[imes][ilag] /= (nanos-1)
                self.FAC[imes][ilag] /= (desvio[imes]*desvio[mes_ant])

        # Calcula funcao de auto-correlacao parcial (uma para cada mes)
        self.FACP = np.zeros((12, ord_max+1), 'd')
        for ilag in np.arange(1,ord_max+1):
            for imes in range(12):
                A = np.eye(ilag)
                B = np.zeros(ilag)
                # Preenche matriz triangular superior
                for ilin in range(len(A)):
                    for icol in range( len(A) ):           # TODO: Aqui poderia ser np.arange(ilin+1,len(A)): Testar depois
                        if icol > ilin:
                            mes = imes - ilin - 1
                            if mes < 0:
                               mes = mes + 12
                            A[ilin][icol] = self.FAC[mes][icol-ilin]
                    B[ilin] = self.FAC[imes][ilin+1]
                # Preenche matriz triangular inferior
                for ilin in range(len(A)):
                    for icol in range( len(A) ):          # TODO: Aqui poderia ser np.arange(0, ilin): Testar depois
                        if icol < ilin:
                            A[ilin][icol] = A[icol][ilin]
                phi = np.linalg.solve(A,B)
                self.FACP[imes][ilag] = phi[ len(phi)-1 ]

        # Identificacao da ordem
        IC = 1.96/sqrt(nanos-1)
        self.Ordem = np.zeros(12, 'i')
        for imes in range(12):
            self.Ordem[imes] = 0
            for ilag in range(ord_max+1):
                if self.FACP[imes][ilag] > IC or self.FACP[imes][ilag] < -IC:
                    self.Ordem[imes] = ilag

        # Calculo dos coeficientes
        self.CoefParp = np.zeros( (12,ord_max), 'd')
        for imes in range(12):
            ilag = self.Ordem[imes]
            A = np.eye(ilag)
            B = np.zeros(ilag)
            # Preenche matriz triangular superior
            for ilin in range(len(A)):
                for icol in range( len(A) ):             # TODO: Aqui poderia ser np.arange(ilin+1,len(A)): Testar depois
                    if icol > ilin:
                        mes = imes - ilin - 1
                        if mes < 0:
                           mes = mes + 12
                        A[ilin][icol] = self.FAC[mes][icol-ilin]
                B[ilin] = self.FAC[imes][ilin+1]
            # Preenche matriz triangular inferior
            for ilin in range(len(A)):
                for icol in range( len(A) ):             # TODO: Aqui poderia ser np.arange(0, ilin): Testar depois
                    if icol < ilin:
                        A[ilin][icol] = A[icol][ilin]
            phi = np.linalg.solve(A,B)
            for iord in range ( len(phi) ):
                self.CoefParp[imes][iord ] = phi[ iord ]

    def gera_series_aditivo(self):

        nanos = len(self.Vazoes) - 1
        ord_max = len(self.CoefParp[0])

        media = np.mean(self.Vazoes[1:nanos], 0)
        desvio = np.std(self.Vazoes[1:nanos], 0)

        # Calculo dos residuos
        residuos = np.zeros( (nanos, 12) )
        for iano in np.arange(1,nanos):
            for imes in range(12):
                residuos[iano][imes]= ( self.Vazoes[iano][imes]-media[imes] ) / desvio[imes]
                for ilag in range(ord_max):
                    ano_ant = iano
                    mes_ant = imes - ilag - 1
                    if mes_ant < 0:
                        ano_ant -= 1
                        mes_ant += 12
                    residuos[iano][imes] -= self.CoefParp[imes][ilag]*( self.Vazoes[ano_ant][mes_ant]-media[mes_ant] ) / desvio[mes_ant]

        # Gera series sinteticas
        sintetica_adit = np.zeros((1000,60),'d')
        for iser in range(1000):
            contador = -1
            for iano in range(5):
                for imes in range(12):
                    contador += 1
                    serie = randint(1,nanos-1)
                    valor = media[imes] + desvio[imes]*residuos[serie][imes]
                    for ilag in range(ord_max):
                        mes_ant = imes - ilag - 1
                        ano_ant = iano
                        if mes_ant < 0:
                            mes_ant += 12
                            ano_ant -= 1
                        if ano_ant < 0:
                            vazant = media[mes_ant]
                        else:
                            vazant = sintetica_adit[iser][contador-1-ilag]
                        valor += desvio[imes]*self.CoefParp[imes][ilag]*(vazant-media[mes_ant])/desvio[mes_ant]
                    sintetica_adit[iser][contador] = valor

        x_axis = np.arange(1, 61)
        plt.plot(x_axis, sintetica_adit.transpose(), 'c-')
        plt.plot(x_axis, np.mean(sintetica_adit,0), 'r-', lw=3, label='Mean - Synthetic Series')
        plt.plot(x_axis, np.mean(sintetica_adit,0) + np.nanstd(sintetica_adit, axis=0), 'r-.', lw=2, label='Std Synthetic Series')
        plt.plot(x_axis, np.mean(sintetica_adit,0) - np.nanstd(sintetica_adit, axis=0), 'r-.', lw=2)
        m = np.concatenate([ media, media, media, media, media])
        d = np.concatenate([ desvio, desvio, desvio, desvio, desvio])
        plt.plot(x_axis, m, 'mo', lw=3, label='Mean - Hystorical Series')
        plt.plot(x_axis, m + d, 'bo', lw=2, label='Std - Hystorical Series')
        plt.plot(x_axis, m - d, 'bo', lw=2)
        titulo = self.Nome.strip() + "'s Synthetic Series of Natural \n" " Inflows - Aditive Noise "
        plt.title(titulo, fontsize=16)
        plt.xlabel('Month', fontsize=16)
        plt.ylabel('Inflow (m^3/s', fontsize=16)
        plt.legend(fontsize=12)
        plt.show()


    def gera_series_multiplicativo(self):

        nanos = len(self.Vazoes) - 1
        ord_max = len(self.CoefParp[0])

        media = np.mean(self.Vazoes[1:nanos], 0)
        desvio = np.std(self.Vazoes[1:nanos], 0)

        # Calculo dos residuos
        residuosmult = np.zeros( (nanos, 12) )
        termoind = np.zeros(12, 'd')
        for iano in np.arange(1,nanos):
            for imes in range(12):
                residuosmult[iano][imes]= self.Vazoes[iano][imes]
                somatorio = 0
                termoind[imes] = media[imes]  # Versao centrada: ver pagina 20 dissertacao Filipe Goulart Cabral (COPPE 2016)
                                              # O ideal portanto, seria utilizar este metodo com formulacao de otimzicao ao
                                              # inves de yule-walker. Restringindo que o termo-ind e os phis sejam todos positivos
                for ilag in range(ord_max):
                    ano_ant = iano
                    mes_ant = imes - ilag - 1
                    if mes_ant < 0:
                        ano_ant -= 1
                        mes_ant += 12
                    somatorio += self.CoefParp[imes][ilag]*self.Vazoes[ano_ant][mes_ant]
                    termoind[imes] -= self.CoefParp[imes][ilag]*media[mes_ant]
                residuosmult[iano][imes] = residuosmult[iano][imes]/(termoind[imes]+somatorio)

        # Gera series sinteticas
        sintetica_mult = np.zeros((1000,60),'d')
        for iser in range(1000):
            contador = -1
            for iano in range(5):
                for imes in range(12):
                    contador += 1
                    serie = randint(1,nanos-1)
                    valor = termoind[imes]
                    for ilag in range(ord_max):
                        mes_ant = imes - ilag - 1
                        ano_ant = iano
                        if mes_ant < 0:
                            mes_ant += 12
                            ano_ant -= 1
                        if ano_ant < 0:
                            vazant = media[mes_ant]
                        else:
                            vazant = sintetica_mult[iser][contador-1-ilag]
                        valor += self.CoefParp[imes][ilag]*vazant
                    sintetica_mult[iser][contador] = valor*residuosmult[serie][imes]

        x_axis = np.arange(1, 61)
        plt.plot(x_axis, sintetica_mult.transpose(), 'c-')
        plt.plot(x_axis, np.mean(sintetica_mult,0), 'r-', lw=3, label='Mean - Synthetic Series')
        plt.plot(x_axis, np.mean(sintetica_mult,0) + np.std(sintetica_mult, axis=0), 'r-.', lw=2, label='Std Synthetic Series')
        plt.plot(x_axis, np.mean(sintetica_mult,0) - np.std(sintetica_mult, axis=0), 'r-.', lw=2)
        m = np.concatenate([ media, media, media, media, media])
        d = np.concatenate([ desvio, desvio, desvio, desvio, desvio])
        plt.plot(x_axis, m, 'mo', lw=3, label='Mean - Hystorical Series')
        plt.plot(x_axis, m + d, 'bo', lw=2, label='Std - Hystorical Series')
        plt.plot(x_axis, m - d, 'bo', lw=2)
        titulo = self.Nome.strip() + "'s Synthetic Series of Natural \n" " Inflows - Multiplicative Noise "
        plt.title(titulo, fontsize=16)
        plt.xlabel('Month', fontsize=16)
        plt.ylabel('Inflows (m3/s)', fontsize=16)
        #plt.ylim(-100, 30000)
        plt.legend(fontsize=12)
        plt.show()

        vasco = 1000

    def pso(self, ord_max):

        def objetivo(x, ord_max, imes):

            nanos = len(self.Vazoes) - 1  # A serie historica do ultimo ano geralmente nao vem completa (despreze-a)

            coef = np.zeros( ord_max+1, 'd')
            for icoef in range(ord_max+1):
                coef[icoef] = x[icoef]

            objetivo = 0.
            residuos = np.zeros(nanos-1)

            for iano in np.arange(1,nanos):
                somatorio = coef[0]
                for ilag in range(ord_max):
                    mes_ant = imes - 1 - ilag
                    ano_ant = iano
                    if mes_ant < 0:
                        mes_ant += 12
                        ano_ant -= 1
                    somatorio += coef[ilag + 1] * self.Vazoes[ano_ant][mes_ant]
                residuos[iano-1]=(self.Vazoes[iano][imes] / somatorio)
                objetivo += ((self.Vazoes[iano][imes] / somatorio) ** 2)

            total = np.sum(residuos) - nanos + 1
            total = total ** 2
            total = 10000*total

            return ( objetivo/(nanos-1) + total )

        self.CoefParp = np.zeros((12, ord_max), 'd')
        self.CoefIndParp = np.zeros(12, 'd')
        self.Ordem = np.zeros(12,'d')
        for imes in range(12):
            print( '*******', imes+1)

            best = 999999
            best_ordem = 0
            for iord in np.arange(1,(ord_max+1)):
                # Define limites e condicao inicial
                lb = np.zeros(iord + 1)
                ub = np.zeros(iord + 1)
                for i in range(iord + 1):
                    lb[i] = 0.00001
                    ub[i] = 10000

                print('**', iord)
                solution, fopt = pso(objetivo, lb, ub, args=(iord, imes), maxiter = 500)

                #solution = minimize(objetivo, x0, method= 'SLSQP', bounds=limites, constraints=cons, args=( iord, imes), options={ 'disp': False, 'maxiter': 10000 }  )
                if fopt < best:
                    best = fopt
                    best_ordem = iord

            self.Ordem[imes] = best_ordem

            solution, fopt = pso(objetivo, lb, ub, args=(iord, imes), maxiter=5000)

            for icoef in range(ord_max):
                self.CoefParp[imes][icoef] = solution[icoef+1]
            self.CoefIndParp[imes] = solution[0]


    def parp_otimo(self, ord_max):


        # Funcao objetivo minimizar erro quatratico medio (o residuo eh o erro)
        def objetivo(x, ord_max, imes):

            nanos = len(self.Vazoes) - 1  # A serie historica do ultimo ano geralmente nao vem completa (despreze-a)

            coef = np.zeros( ord_max+1, 'd')
            for icoef in range(ord_max+1):
                coef[icoef] = x[icoef]

            objetivo = 0.
            for iano in np.arange(1,nanos):
                somatorio = coef[0]
                for ilag in range(ord_max):
                    mes_ant = imes - 1 - ilag
                    ano_ant = iano
                    if mes_ant < 0:
                        mes_ant += 12
                        ano_ant -= 1
                    somatorio += coef[ilag + 1] * self.Vazoes[ano_ant][mes_ant]
                objetivo += ((self.Vazoes[iano][imes] / somatorio) ** 2)

            return objetivo/(nanos-1)

        # Restricoes a media dos residuos devem ser igual a unidade ou somatorio dos residuos devem ser igual a nanos
        def restricao(x, ord_max, imes):

            nanos = len(self.Vazoes) - 1  # A serie historica do ultimo ano geralmente nao vem completa (despreze-a)

            coef = np.zeros( ord_max+1, 'd')
            for icoef in range(ord_max+1):
                coef[icoef] = x[icoef]

            objetivo = 0.
            residuos = np.zeros(nanos-1)
            for iano in np.arange(1,nanos):
                somatorio = coef[0]
                for ilag in range(ord_max):
                    mes_ant = imes - 1 - ilag
                    ano_ant = iano
                    if mes_ant < 0:
                        mes_ant += 12
                        ano_ant -= 1
                    somatorio += coef[ilag + 1] * self.Vazoes[ano_ant][mes_ant]
                objetivo += (self.Vazoes[iano][imes] / somatorio)
                residuos[iano-1]=(self.Vazoes[iano][imes] / somatorio)
                desvio = np.std(residuos)
                #curtose = kurtosis(residuos) - 3
            # return ([ objetivo - nanos + 1 , desvio - 0.2])
            return ([ objetivo - nanos + 1 ])

        # Define limites e condicao inicial
        x0 = np.zeros(ord_max+1)
        limites = []
        for i in range(ord_max + 1):
            x0[i] = 0.1
            limites.append((0, 10000))

        self.CoefParp = np.zeros((12, ord_max), 'd')
        self.CoefIndParp = np.zeros(12, 'd')
        self.Ordem = np.zeros(12,'d')
        for imes in range(12):
            print( '*******', imes+1)
            conl = {'type': 'eq', 'fun': restricao, 'args': (ord_max, imes)}
            cons = ([conl])
            best = 999999
            best_ordem = 0
            #for iord in np.arange(1,(ord_max+1)):
            for iord in np.arange(6, 7):
                solution = minimize(objetivo, x0, method= 'SLSQP', bounds=limites, constraints=cons, args=( iord, imes), options={ 'disp': False, 'maxiter': 10000 }  )
                if solution.fun < best:
                    best = solution.fun
                    best_ordem = iord

            self.Ordem[imes] = best_ordem
            solution = minimize(objetivo, x0, method= 'SLSQP', bounds=limites, constraints=cons, args=( best_ordem, imes), options={ 'disp': True, 'maxiter': 10000 }  )


            for icoef in range(ord_max):
                self.CoefParp[imes][icoef] = solution.x[icoef+1]
            self.CoefIndParp[imes] = solution.x[0]


    def gera_series_multiplicativo_parp_otimo(self):

        nanos = len(self.Vazoes) - 1
        ord_max = len(self.CoefParp[0])

        media = np.mean(self.Vazoes[1:nanos], 0)
        desvio = np.std(self.Vazoes[1:nanos], 0)

        # Calculo dos residuos
        residuosmult = np.zeros( (nanos, 12) )
        for iano in np.arange(1,nanos):
            for imes in range(12):
                residuosmult[iano][imes]= self.Vazoes[iano][imes]
                somatorio = 0
                for ilag in range(ord_max):
                    ano_ant = iano
                    mes_ant = imes - ilag - 1
                    if mes_ant < 0:
                        ano_ant -= 1
                        mes_ant += 12
                    somatorio += self.CoefParp[imes][ilag]*self.Vazoes[ano_ant][mes_ant]
                residuosmult[iano][imes] = residuosmult[iano][imes]/(self.CoefIndParp[imes]+somatorio)

        # Gera series sinteticas
        sintetica_mult = np.zeros((1000,60),'d')
        for iser in range(1000):
            contador = -1
            for iano in range(5):
                for imes in range(12):
                    contador += 1
                    serie = randint(1,nanos-1)
                    valor = self.CoefIndParp[imes]
                    for ilag in range(ord_max):
                        mes_ant = imes - ilag - 1
                        ano_ant = iano
                        if mes_ant < 0:
                            mes_ant += 12
                            ano_ant -= 1
                        if ano_ant < 0:
                            vazant = media[mes_ant]
                        else:
                            vazant = sintetica_mult[iser][contador-1-ilag]
                        valor += self.CoefParp[imes][ilag]*vazant
                    sintetica_mult[iser][contador] = valor*residuosmult[serie][imes]

        x_axis = np.arange(1, 61)
        plt.plot(x_axis, sintetica_mult.transpose(), 'c-')
        plt.plot(x_axis, np.mean(sintetica_mult,0), 'r-', lw=3, label='Mean - Synthetic Series')
        plt.plot(x_axis, np.mean(sintetica_mult,0) + np.std(sintetica_mult, axis=0), 'r-.', lw=2, label='Std Synthetic Series')
        plt.plot(x_axis, np.mean(sintetica_mult,0) - np.std(sintetica_mult, axis=0), 'r-.', lw=2)
        m = np.concatenate([ media, media, media, media, media])
        d = np.concatenate([ desvio, desvio, desvio, desvio, desvio])
        plt.plot(x_axis, m, 'mo', lw=3, label='Mean - Hystorical Series')
        plt.plot(x_axis, m + d, 'bo', lw=2, label='Std - Hystorical Series')
        plt.plot(x_axis, m - d, 'bo', lw=2)
        titulo = self.Nome.strip() + "'s Synthetic Series of Natural \n" " Inflows - Multiplicative Noise "
        plt.title(titulo, fontsize=16)
        plt.xlabel('Month', fontsize=16)
        plt.ylabel('Inflows (m3/s)', fontsize=16)
        #plt.ylim(-100, 30000)
        plt.legend(fontsize=12)
        plt.show()

        vasco = 1000



import struct
from src import hidr
from src import term
from src import resequiv
from src import submercado as subsist
from src import interc
import numpy as np

class pmo(object):

    diretorio = None
    nome_arquivo    = None
    nome_dger       = None
    nome_sistema    = None
    nome_confhd     = None
    nome_modif      = None
    nome_conft      = None
    nome_term       = None
    nome_clast      = None
    nome_exph       = None
    nome_expt       = None
    nome_patamar    = None
    nome_cortes     = None
    nome_corteh     = None
    nome_pmo        = None
    nome_parp       = None
    nome_forward    = None
    nome_fowardh    = None
    nome_shist      = None
    nome_manutt     = None
    nome_newdesp    = None
    nome_vazpast    = None
    nome_itaipu     = None
    nome_bid        = None
    nome_c_adic     = None
    nome_loss       = None
    nome_gtminpat   = None
    nome_elnino     = None
    nome_ensoaux    = None
    nome_dsvagua    = None
    nome_penalid    = None
    nome_curva      = None
    nome_agrint     = None
    nome_adterm     = None
    nome_ghmin      = None
    nome_sar        = None
    nome_cvar       = None
    nome_ree        = None
    nome_re         = None
    nome_hidr       = 'HIDR.DAT'
    nome_vazoes     = 'VAZOES.DAT'

    def __init__(self, diretorio):
        self.diretorio = diretorio

    def le_caso(self):

        # Le nome de arquivo com lista de arquivos de entrada
        file_name = self.diretorio + 'CASO.DAT'
        file = open(file_name, "r")
        arquivo = file.readlines()
        file.close()
        self.nome_arquivo = arquivo[0][0:len(arquivo[0])-1]

        # Le nomes de todos os arquivos de entrada e saida do modelo NEWAVE
        file_name = self.diretorio + self.nome_arquivo
        file = open(file_name, "r")
        arquivo = file.readlines()
        file.close()

        self.nome_dger      = arquivo[0][30:len(arquivo[0])-1]
        self.nome_sistema   = arquivo[1][30:len(arquivo[1])-1]
        self.nome_confhd    = arquivo[2][30:len(arquivo[2])-1]
        self.nome_modif     = arquivo[3][30:len(arquivo[3])-1]
        self.nome_conft     = arquivo[4][30:len(arquivo[4])-1]
        self.nome_term      = arquivo[5][30:len(arquivo[5])-1]
        self.nome_clast     = arquivo[6][30:len(arquivo[6])-1]
        self.nome_exph      = arquivo[7][30:len(arquivo[7])-1]
        self.nome_expt      = arquivo[8][30:len(arquivo[8])-1]
        self.nome_patamar   = arquivo[9][30:len(arquivo[9])-1]
        self.nome_cortes    = arquivo[10][30:len(arquivo[10])-1]
        self.nome_corteh    = arquivo[11][30:len(arquivo[11])-1]
        self.nome_pmo       = arquivo[12][30:len(arquivo[12])-1]
        self.nome_parp      = arquivo[13][30:len(arquivo[13])-1]
        self.nome_forward   = arquivo[14][30:len(arquivo[14])-1]
        self.nome_fowardh   = arquivo[15][30:len(arquivo[15])-1]
        self.nome_shist     = arquivo[16][30:len(arquivo[16])-1]
        self.nome_manutt    = arquivo[17][30:len(arquivo[17])-1]
        self.nome_newdesp   = arquivo[18][30:len(arquivo[18])-1]
        self.nome_vazpast   = arquivo[19][30:len(arquivo[19])-1]
        self.nome_itaipu    = arquivo[20][30:len(arquivo[20])-1]
        self.nome_bid       = arquivo[21][30:len(arquivo[21])-1]
        self.nome_c_adic    = arquivo[22][30:len(arquivo[22])-1]
        self.nome_loss      = arquivo[23][30:len(arquivo[23])-1]
        self.nome_gtminpat  = arquivo[24][30:len(arquivo[24])-1]
        self.nome_elnino    = arquivo[25][30:len(arquivo[25])-1]
        self.nome_ensoaux   = arquivo[26][30:len(arquivo[26])-1]
        self.nome_dsvagua   = arquivo[27][30:len(arquivo[27])-1]
        self.nome_penalid   = arquivo[28][30:len(arquivo[28])-1]
        self.nome_curva     = arquivo[29][30:len(arquivo[29])-1]
        self.nome_agrint    = arquivo[30][30:len(arquivo[30])-1]
        self.nome_adterm    = arquivo[31][30:len(arquivo[31])-1]
        self.nome_ghmin     = arquivo[32][30:len(arquivo[32])-1]
        self.nome_sar       = arquivo[33][30:len(arquivo[33])-1]
        self.nome_cvar      = arquivo[35][30:len(arquivo[34])-1]
        self.nome_ree       = arquivo[35][30:len(arquivo[35])-1]
        self.nome_re        = arquivo[36][30:len(arquivo[36])-1]
        print('OK! Leitura do CASO.DAT e ARQUIVOS.DAT realizada com sucesso.')

    def le_hidr(self, cadastro):

        file_name = self.diretorio + self.nome_hidr
        file = open(file_name, "rb")
        nreg = 320

        i = 0
        while i < nreg:
            cadastro.append(hidr())
            iusi = len(cadastro) - 1
            cadastro[iusi].Codigo = i+1
            cadastro[iusi].Nome = struct.unpack('12s', file.read(12))[0]
            cadastro[iusi].Posto = struct.unpack('i', file.read(4))[0]
            cadastro[iusi].Bdh = struct.unpack('8s', file.read(8))[0]
            cadastro[iusi].Sist = struct.unpack('i', file.read(4))[0]
            cadastro[iusi].Empr = struct.unpack('i', file.read(4))[0]
            cadastro[iusi].Jusante = struct.unpack('i', file.read(4))[0]
            cadastro[iusi].Desvio = struct.unpack('i', file.read(4))[0]
            cadastro[iusi].VolMin = struct.unpack('f', file.read(4))[0]
            cadastro[iusi].VolMax = struct.unpack('f', file.read(4))[0]
            cadastro[iusi].VolVert = struct.unpack('f', file.read(4))[0]
            cadastro[iusi].VolMinDesv = struct.unpack('f', file.read(4))[0]
            cadastro[iusi].CotaMin = struct.unpack('f', file.read(4))[0]
            cadastro[iusi].CotaMax = struct.unpack('f', file.read(4))[0]
            cadastro[iusi].PolCotaVol = list(struct.unpack('5f', bytearray(file.read(20))))
            cadastro[iusi].PolCotaArea = list(struct.unpack('5f', bytearray(file.read(20))))
            cadastro[iusi].CoefEvap = list(struct.unpack('12i', bytearray(file.read(48))))
            cadastro[iusi].NumConjMaq = struct.unpack('i', file.read(4))[0]
            cadastro[iusi].MaqporConj = list(struct.unpack('5i', bytearray(file.read(20))))
            cadastro[iusi].PEfporConj = list(struct.unpack('5f', bytearray(file.read(20))))

            cadastro[iusi].CF_HBQT = []
            cadastro[iusi].CF_HBQG = []
            cadastro[iusi].CF_HBPT = []
            for j in range(5):
                cadastro[iusi].CF_HBQT.append(list(struct.unpack('5f', bytearray(file.read(20)))))
            for j in range(5):
                cadastro[iusi].CF_HBQG.append(list(struct.unpack('5f', bytearray(file.read(20)))))
            for j in range(5):
                cadastro[iusi].CF_HBPT.append(list(struct.unpack('5f', bytearray(file.read(20)))))

            cadastro[iusi].AltEfetConj = list(struct.unpack('5f', bytearray(file.read(20))))
            cadastro[iusi].VazEfetConj = list(struct.unpack('5i', bytearray(file.read(20))))
            cadastro[iusi].ProdEsp = struct.unpack('f', file.read(4))[0]
            cadastro[iusi].PerdaHid = struct.unpack('f', file.read(4))[0]
            cadastro[iusi].NumPolVNJ = struct.unpack('i', file.read(4))[0]

            cadastro[iusi].PolVazNivJus = []
            for j in range(5):
                cadastro[iusi].PolVazNivJus.append(list(struct.unpack('6f', bytearray(file.read(24)))))

            cadastro[iusi].CotaRefNivelJus = list(struct.unpack('6f', bytearray(file.read(24))))
            cadastro[iusi].CFMed = struct.unpack('f', file.read(4))[0]
            cadastro[iusi].InfCanalFuga = struct.unpack('i', file.read(4))[0]
            cadastro[iusi].FatorCargaMax = struct.unpack('f', file.read(4))[0]
            cadastro[iusi].FatorCargaMin = struct.unpack('f', file.read(4))[0]
            cadastro[iusi].VazMin = struct.unpack('i', file.read(4))[0]
            cadastro[iusi].UnidBase = struct.unpack('i', file.read(4))[0]
            cadastro[iusi].TipoTurb = struct.unpack('i', file.read(4))[0]
            cadastro[iusi].RepresConj = struct.unpack('i', file.read(4))[0]
            cadastro[iusi].TEIF = struct.unpack('f', file.read(4))[0]
            cadastro[iusi].IP = struct.unpack('f', file.read(4))[0]
            cadastro[iusi].TipoPerda = struct.unpack('i', file.read(4))[0]
            cadastro[iusi].Data = struct.unpack('8s', file.read(8))[0]
            cadastro[iusi].Observ = struct.unpack('43s', file.read(43))[0]
            cadastro[iusi].VolRef = struct.unpack('f', file.read(4))[0]
            cadastro[iusi].TipoReg = struct.unpack('c', file.read(1))[0]

            i = i + 1

        file.close()
        print('OK! Leitura do HIDR.DAT realizada com sucesso. (', nreg, 'Usinas Hidraulicas )')

        return cadastro

    def le_confh(self, conf, cadastro, nanos):

        file_name = self.diretorio + self.nome_confhd
        file = open(file_name, "r")
        arquivo = file.readlines()
        file.close()

        vazoes_name = self.diretorio + self.nome_vazoes
        file_vazoes = open(vazoes_name , 'rb')
        vazoes = np.fromfile(file_vazoes, dtype=np.int32)
        file_vazoes.close()
        # A principio o numero de anos seria ano corrente - 2, mas
        # o ons tem usado complementar o arquivo de vazoes com os
        # dados ate o mes anterior do ano corrente
        # Calcular o numero de anos dividindo o tamanho da variavel data / nreg / nmeses

        num_anos = int(vazoes.shape[0] / 320 / 12)
        vaz_nat = vazoes.reshape(num_anos*12, 320)

        for i in range(2,len(arquivo)):
            if len(arquivo[i])>5:
                codigo = int(arquivo[i][1:5])
            else:
                break
            for iusi, usina in enumerate(cadastro):
                if usina.Codigo == codigo:
                    conf.append(usina)
                    indice = len(conf)-1
                    conf[indice].Nome = arquivo[i][6:18]
                    conf[indice].Posto = int(arquivo[i][19:23])
                    conf[indice].Jusante = int(arquivo[i][25:29])
                    conf[indice].Ree = int(arquivo[i][30:34])
                    conf[indice].VolIni = float(arquivo[i][35:41])
                    conf[indice].Status = arquivo[i][44:46]
                    conf[indice].Modif = int(arquivo[i][49:53])
                    conf[indice].AnoI = int(arquivo[i][58:62])
                    conf[indice].AnoF = int(arquivo[i][67:71])

                    # Se a usina for 'NE' ou 'EE' nao deve possuir maquinas
                    if conf[indice].Status == 'NE' or conf[indice].Status == 'EE':
                        for iconj in range(5):
                            conf[indice].MaqporConj[iconj] = 0

                    # Parametros Temporais controlados pelo MODIF.DAT
                    conf[indice].VolMinT = conf[indice].VolMin * np.ones((nanos,12), 'f')
                    conf[indice].VolMaxT = conf[indice].VolMax * np.ones((nanos,12), 'f')
                    conf[indice].VolMinP = conf[indice].VolMin * np.ones((nanos,12), 'f')
                    conf[indice].VazMinT = conf[indice].VazMin * np.ones((nanos,12), 'f')
                    conf[indice].CFugaT  = conf[indice].CFMed  * np.ones((nanos,12), 'f')

                    # Calcula Parametros
                    conf[indice].CalcVolUtil()
                    conf[indice].CalcVazEfetiva()
                    conf[indice].CalcPotEfetiva()
                    conf[indice].CalcProdutibs(nanos)
                    conf[indice].CalcEngolMaximo()

                    # Parametros Temporais calculados pelo EXPH.DAT
                    if conf[indice].Status == 'EX':
                        conf[indice].StatusVolMorto = 2*np.ones((nanos,12),'i')
                        conf[indice].StatusMotoriz = 2*np.ones((nanos,12),'i')
                        conf[indice].VolMortoTempo = np.zeros((nanos,12),'f')
                        conf[indice].EngolTempo = conf[indice].Engolimento * np.ones((nanos,12),'f')
                        conf[indice].PotenciaTempo = conf[indice].PotEfet * np.ones((nanos,12),'f')
                        conf[indice].UnidadesTempo = sum(conf[indice].MaqporConj)*np.ones((nanos,12),'f')
                    else:
                        conf[indice].StatusVolMorto = np.zeros((nanos,12),'i')
                        conf[indice].StatusMotoriz = np.zeros((nanos,12),'i')
                        conf[indice].VolMortoTempo = np.zeros((nanos, 12), 'f')
                        conf[indice].UnidadesTempo = np.zeros((nanos, 12), 'i')
                        if conf[indice].Status == 'EE':
                            conf[indice].EngolTempo = conf[indice].Engolimento * np.ones((nanos,12),'f')
                            conf[indice].PotenciaTempo = conf[indice].PotEfet * np.ones((nanos,12),'f')
                        else:
                            conf[indice].EngolTempo = np.zeros((nanos,12),'f')
                            conf[indice].PotenciaTempo = np.zeros((nanos,12),'f')

                    # Le Historico de Vazoes Naturais da Usina
                    posto = conf[indice].Posto-1
                    vaz_nat_usina = np.zeros(num_anos*12, int)
                    for i in range(num_anos*12):
                        vaz_nat_usina[i] = vaz_nat[i][posto]
                    conf[indice].Vazoes = vaz_nat_usina
                    conf[indice].Vazoes.shape = (num_anos,12)

                    break

        print('OK! Leitura do CONFHD e VAZOES realizada com sucesso. (', indice+1, 'Usinas Hidraulicas )')
        return conf

    def le_term(self, cadastro):

        file_name = self.diretorio + self.nome_term
        file = open(file_name, "r")
        arquivo = file.readlines()
        file.close()

        for i in range(2,len(arquivo)):
            if len(arquivo[i])>134:
                cadastro.append(term())
                indice = len(cadastro) - 1
                cadastro[indice].Codigo     = int(arquivo[i][1:4])
                cadastro[indice].Nome       = arquivo[i][5:17]
                cadastro[indice].Potencia   = float(arquivo[i][19:24])
                cadastro[indice].FCMax      = float(arquivo[i][25:29])
                cadastro[indice].TEIF       = float(arquivo[i][31:37])
                cadastro[indice].IP         = float(arquivo[i][38:44])
                gtmin = np.zeros(13,float)
                for j in range(13):
                    gtmin[j] = float(arquivo[i][(45+j*7):(51+j*7)])
                cadastro[indice].GTMin      = gtmin
            else:
                break


        print('OK! Leitura do TERM realizada com sucesso. (', indice+1, 'Usinas Termicas )' )
        return cadastro

    def le_conft(self, conf, cadastro):

        file_name = self.diretorio + self.nome_conft
        file = open(file_name, "r")
        arquivo = file.readlines()
        file.close()

        for i in range(2,len(arquivo)):
            if len(arquivo[i])>5:
                codigo = int(arquivo[i][2:5])
            else:
                break
            encontrei = False
            for iusi, usina in enumerate(cadastro):
                if usina.Codigo == codigo:
                    encontrei = True
                    conf.append(usina)
                    indice = len(conf)-1
                    conf[indice].Nome = arquivo[i][6:18]
                    conf[indice].Sist = int(arquivo[i][21:25])
                    conf[indice].Status = arquivo[i][30:32]
                    conf[indice].Classe = int(arquivo[i][35:39])
                    break
            if not encontrei:
                print ('Ops.... Usina', codigo, 'do CONFT nao encontrada no TERM')

        print('OK! Leitura do CONFT realizada com sucesso. (', indice+1, 'Usinas Termicas )')

        return conf

    def le_clast(self, conf):

        file_name = self.diretorio + self.nome_clast
        file = open(file_name, "r")
        arquivo = file.readlines()
        file.close()

        contador = 0
        for i in range(2,len(arquivo)):
            if len(arquivo[i])>5:
                codigo = int(arquivo[i][1:5])
                if codigo == 9999:
                    break
            else:
                break
            for iusi, usina in enumerate(conf):
                if usina.Classe == codigo:
                    contador += 1
                    usina.NomeClasse = arquivo[i][6:18]
                    usina.TipoComb = arquivo[i][19:29]
                    custo = np.zeros(5, float)
                    for j in range(5):
                        custo[j] = float(arquivo[i][(30 + j * 8):(37 + j * 8)])
                    usina.Custo = custo
                    break

        print('OK! Leitura do CLAST realizada com sucesso. (', contador, 'Usinas Termicas )')

        return conf

    def le_modif(self, conf, anoinicial, nanos):
        file_name = self.diretorio + self.nome_modif
        file = open(file_name, "r")
        arquivo = file.readlines()
        file.close()

        i = 2
        nr_modificacoes = 0
        while i < len(arquivo)-1:
            codigo = int(arquivo[i][6:13])
            encontrou = False
            for iusi, usina in enumerate(conf):
                if usina.Codigo == codigo:
                    nr_modificacoes += 1
                    encontrou = True
                    break
            i=i+1
            if not encontrou:
                print ('Usina com codigo', codigo, 'listada no modif nao encontrada.')
                while arquivo[i][1:6] != 'USINA' and i < len(arquivo)-2:
                    i = i + 1
            while arquivo[i][1:6] != 'USINA' and encontrou:
                if   arquivo[i][1:7] == 'VOLMIN'    or arquivo[i][1:7] == 'volmin':
                    j = 7
                    texto = arquivo[i][j:len(arquivo[i]) - 1]
                    texto = texto.split(" ")
                    while '' in texto:
                        texto.remove('')
                    if texto[1] == "'%'":
                        usina.VolMin = usina.VolMin + float(texto[0])*usina.VolUtil/100
                    else:
                        usina.VolMin = float(texto[0])
                    usina.CalcVolUtil()
                    usina.CalcProdutibs(nanos)
                    usina.CalcEngolMaximo()
                    usina.VolMinT = usina.VolMin * np.ones((nanos,12), 'f')
                    usina.VolMinP = usina.VolMin * np.ones((nanos,12), 'f')
                elif arquivo[i][1:7] == 'VOLMAX'    or arquivo[i][1:7] == 'volmax':
                    j = 7
                    texto = arquivo[i][j:len(arquivo[i]) - 1]
                    texto = texto.split(" ")
                    while '' in texto:
                        texto.remove('')
                    if texto[1] == "'%'":
                        usina.VolMax = usina.VolMin + float(texto[0])*usina.VolUtil/100
                    else:
                        usina.VolMax = float(texto[0])
                    usina.CalcVolUtil()
                    usina.CalcProdutibs(nanos)
                    usina.CalcEngolMaximo()
                    usina.VolMaxT = usina.VolMax * np.ones((nanos,12), 'f')
                elif arquivo[i][1:7] == 'NUMCNJ'    or arquivo[i][1:7] == 'numcnj':
                    j = 7
                    texto = arquivo[i][j:len(arquivo[i]) - 1]
                    texto = texto.split(" ")
                    while '' in texto:
                        texto.remove('')
                    usina.NumConjMaq = int(texto[0])
                    usina.CalcVazEfetiva()
                    usina.CalcPotEfetiva()
                    usina.CalcEngolMaximo()
                    usina.UnidadesTempo = sum(usina.MaqporConj) * np.ones((nanos, 12), 'f')
                    if usina.Status == 'EE' or usina.Status == 'EX':
                        usina.EngolTempo = usina.Engolimento * np.ones((nanos, 12), 'f')
                        usina.PotenciaTempo = usina.PotEfet * np.ones((nanos, 12), 'f')
                    else:
                        usina.EngolTempo = np.zeros((nanos, 12), 'f')
                        usina.PotenciaTempo = np.zeros((nanos, 12), 'f')
                elif arquivo[i][1:7] == 'NUMMAQ'    or arquivo[i][1:7] == 'nummaq':
                    j = 7
                    texto = arquivo[i][j:len(arquivo[i]) - 1]
                    texto = texto.split(" ")
                    while '' in texto:
                        texto.remove('')
                    numero = int(texto[1])
                    usina.MaqporConj[numero-1] = int(texto[0])
                    usina.CalcVazEfetiva()
                    usina.CalcPotEfetiva()
                    usina.CalcEngolMaximo()
                    usina.UnidadesTempo = sum(usina.MaqporConj) * np.ones((nanos, 12), 'f')
                    if usina.Status == 'EE' or usina.Status == 'EX':
                        usina.EngolTempo = usina.Engolimento * np.ones((nanos, 12), 'f')
                        usina.PotenciaTempo = usina.PotEfet * np.ones((nanos, 12), 'f')
                    else:
                        usina.EngolTempo = np.zeros((nanos, 12), 'f')
                        usina.PotenciaTempo = np.zeros((nanos, 12), 'f')
                elif arquivo[i][1:7] == 'POTEFE'    or arquivo[i][1:7] == 'potefe':
                    j = 7
                    texto = arquivo[i][j:len(arquivo[i]) - 1]
                    texto = texto.split(" ")
                    while '' in texto:
                        texto.remove('')
                    numero = int(texto[1])
                    usina.PEfporConj[numero-1] = float(texto[0])
                    usina.CalcPotEfetiva()
                    usina.CalcEngolMaximo()
                    if usina.Status == 'EE' or usina.Status == 'EX':
                        usina.EngolTempo = usina.Engolimento * np.ones((nanos, 12), 'f')
                        usina.PotenciaTempo = usina.PotEfet * np.ones((nanos, 12), 'f')
                    else:
                        usina.EngolTempo = np.zeros((nanos, 12), 'f')
                        usina.PotenciaTempo = np.zeros((nanos, 12), 'f')
                elif arquivo[i][1:8] == 'PRODESP'   or arquivo[i][1:8] == 'prodesp':
                    j = 8
                    texto = arquivo[i][j:len(arquivo[i]) - 1]
                    texto = texto.split(" ")
                    while '' in texto:
                        texto.remove('')
                    usina.ProdEsp = float(texto[0])
                    usina.CalcProdutibs(nanos)
                    usina.CalcEngolMaximo()
                    if usina.Status == 'EE' or usina.Status == 'EX':
                        usina.EngolTempo = usina.Engolimento * np.ones((nanos, 12), 'f')
                        usina.PotenciaTempo = usina.PotEfet * np.ones((nanos, 12), 'f')
                    else:
                        usina.EngolTempo = np.zeros((nanos, 12), 'f')
                        usina.PotenciaTempo = np.zeros((nanos, 12), 'f')
                elif arquivo[i][1:5] == 'TEIF'      or arquivo[i][1:5] == 'teif':
                    j = 5
                    texto = arquivo[i][j:len(arquivo[i]) - 1]
                    texto = texto.split(" ")
                    while '' in texto:
                        texto.remove('')
                    usina.TEIF = float(texto[0])
                elif arquivo[i][1:3] == 'IP'        or arquivo[i][1:3] == 'ip':
                    j = 3
                    texto = arquivo[i][j:len(arquivo[i]) - 1]
                    texto = texto.split(" ")
                    while '' in texto:
                        texto.remove('')
                    usina.IP = float(texto[0])
                elif arquivo[i][1:9] == 'PERDHIDR'  or arquivo[i][1:9] == 'perdhidr':
                    j = 9
                    texto = arquivo[i][j:len(arquivo[i]) - 1]
                    texto = texto.split(" ")
                    while '' in texto:
                        texto.remove('')
                    usina.PerdaHid = float(texto[0])
                elif arquivo[i][1:8] == 'VAZMIN '    or arquivo[i][1:8] == 'vazmin ':
                    j = 8
                    texto = arquivo[i][j:len(arquivo[i]) - 1]
                    texto = texto.split(" ")
                    while '' in texto:
                        texto.remove('')
                    usina.VazMin = float(texto[0])
                    usina.VazMinT = usina.VazMin * np.ones((nanos,12), 'f')
                elif arquivo[i][1:9] == 'COEFEVAP'  or arquivo[i][1:9] == 'coefevap':
                    j = 9
                    texto = arquivo[i][j:len(arquivo[i]) - 1]
                    texto = texto.split(" ")
                    while '' in texto:
                        texto.remove('')
                    numero = int(texto[1])
                    usina.CoefEvap[numero] = int(texto[0])
                elif arquivo[i][1:8] == 'COTAREA'   or arquivo[i][1:8] == 'cotarea':
                    j = 8
                    texto = arquivo[i][j:len(arquivo[i]) - 1]
                    texto = texto.split(" ")
                    while '' in texto:
                        texto.remove('')
                    for i in range(5):
                        usina.PolCotaArea[i] = float(texto[i])
                elif arquivo[i][1:8] == 'VOLCOTA'   or arquivo[i][1:8] == 'volcota':
                    j = 8
                    texto = arquivo[i][j:len(arquivo[i]) - 1]
                    texto = texto.split(" ")
                    while '' in texto:
                        texto.remove('')
                    for i in range(5):
                        usina.PolCotaVol[i] = float(texto[i])
                    usina.CalcProdutibs(nanos)
                    usina.CalcEngolMaximo()
                    if usina.Status == 'EE' or usina.Status == 'EX':
                        usina.EngolTempo = usina.Engolimento * np.ones((nanos, 12), 'f')
                        usina.PotenciaTempo = usina.PotEfet * np.ones((nanos, 12), 'f')
                    else:
                        usina.EngolTempo = np.zeros((nanos, 12), 'f')
                        usina.PotenciaTempo = np.zeros((nanos, 12), 'f')
                elif arquivo[i][1:6] == 'CFUGA'     or arquivo[i][1:6] == 'cfuga':
                    j = 6
                    texto = arquivo[i][j:len(arquivo[i]) - 1]
                    texto = texto.split(" ")
                    while '' in texto:
                        texto.remove('')
                    valor = float(texto[2])
                    mesinicial = int(texto[0])
                    for iano in range(int(texto[1])-anoinicial,nanos):
                        for imes in range(mesinicial-1,12):
                            usina.CFugaT[iano][imes] = valor
                        mesinicial = 0
                    usina.CalcProdutibs(nanos)
                    #usina.CFMed = float(texto[0])
                    #
                    #usina.CalcEngolMaximo()
                    #if usina.Status == 'EE' or usina.Status == 'EX':
                    #    usina.EngolTempo = usina.Engolimento * np.ones((nanos, 12), 'f')
                    #    usina.PotenciaTempo = usina.PotEfet * np.ones((nanos, 12), 'f')
                    #else:
                    #    usina.EngolTempo = np.zeros((nanos, 12), 'f')
                    #    usina.PotenciaTempo = np.zeros((nanos, 12), 'f')
                elif arquivo[i][1:6] == 'VMAXT'     or arquivo[i][1:6] == 'vmaxt':
                    j = 6
                    texto = arquivo[i][j:len(arquivo[i]) - 1]
                    texto = texto.split(" ")
                    while '' in texto:
                        texto.remove('')
                    valor = float(texto[2])
                    if (texto[3] == "'%'"):
                        valor = usina.VolMin + usina.VolUtil * float(texto[2])/ 100
                    mesinicial = int(texto[0])
                    for iano in range(int(texto[1])-anoinicial,nanos):
                        for imes in range(mesinicial-1,12):
                            usina.VolMaxT[iano][imes] = valor
                        mesinicial = 0
                elif arquivo[i][1:6] == 'VMINT'     or arquivo[i][1:6] == 'vmint':
                    j = 6
                    texto = arquivo[i][j:len(arquivo[i]) - 1]
                    texto = texto.split(" ")
                    while '' in texto:
                        texto.remove('')
                    valor = float(texto[2])
                    if (texto[3] == "'%'"):
                        valor = usina.VolMin + usina.VolUtil * float(texto[2])/ 100
                    mesinicial = int(texto[0])
                    for iano in range(int(texto[1])-anoinicial,nanos):
                        for imes in range(mesinicial-1,12):
                            usina.VolMinT[iano][imes] = valor
                        mesinicial = 0
                elif arquivo[i][1:7] == 'NUMBAS'    or arquivo[i][1:7] == 'numbas':
                    j = 7
                    texto = arquivo[i][j:len(arquivo[i]) - 1]
                    texto = texto.split(" ")
                    while '' in texto:
                        texto.remove('')
                    usina.UnidBase = int(texto[0])
                elif arquivo[i][1:6] == 'VMINP'     or arquivo[i][1:6] == 'vminp':
                    j = 6
                    texto = arquivo[i][j:len(arquivo[i]) - 1]
                    texto = texto.split(" ")
                    while '' in texto:
                        texto.remove('')
                    valor = float(texto[2])
                    if (texto[3] == "'%'"):
                        valor = usina.VolMin + usina.VolUtil * float(texto[2])/ 100
                    mesinicial = int(texto[0])
                    for iano in range(int(texto[1])-anoinicial,nanos):
                        for imes in range(mesinicial-1,12):
                            usina.VolMinP[iano][imes] = valor
                        mesinicial = 0
                elif arquivo[i][1:8] == 'VAZMINT'   or arquivo[i][1:8] == 'vazmint':
                    j = 8
                    texto = arquivo[i][j:len(arquivo[i]) - 1]
                    texto = texto.split(" ")
                    while '' in texto:
                        texto.remove('')
                    valor = float(texto[2])
                    mesinicial = int(texto[0])
                    for iano in range(int(texto[1])-anoinicial,nanos):
                        for imes in range(mesinicial-1,12):
                            usina.VazMinT[iano][imes] = valor
                        mesinicial = 0
                else:
                    print('Oooopppppsssss: Erro Leitura MODIF.DAT -  Linha', i+1)
                i = i + 1
                if i == len(arquivo):
                    break
        print('OK! Leitura do MODIF realizada com sucesso. (', nr_modificacoes, 'Usinas Hidraulicas Modificadas - ',i,'linhas lidas )')
        return conf

    def le_exph(self, conf, anoinicial, nanos):
        file_name = self.diretorio + self.nome_exph
        file = open(file_name, "r")
        arquivo = file.readlines()
        file.close()

        i = 3
        nr_expansoes = 0
        while i < len(arquivo) and len(arquivo[i]) > 3:
            codigo = int(arquivo[i][0:4])
            for iusi, usina in enumerate(conf):
                if usina.Codigo == codigo:
                    if arquivo[i][18:20] != '  ':
                        mes_vm = int(arquivo[i][18:20])
                    else:
                        mes_vm = 0
                    if arquivo[i][21:25] != '    ':
                        ano_vm = int(arquivo[i][21:25])
                    else:
                        ano_vm = 0
                    if arquivo[i][31:33] != '  ':
                        dur_vm = int(arquivo[i][31:33])
                    else:
                        dur_vm = 0
                    if arquivo[i][38:42] != '    ':
                        per_vm = float(arquivo[i][38:42])
                    else:
                        per_vm = 0.

                    if len(arquivo[i]) > 64:
                        mes_ent  = int(arquivo[i][44:46])
                        ano_ent  = int(arquivo[i][47:51])
                        pot_ent  = float(arquivo[i][52:58])
                        unidade  = int(arquivo[i][60:62])
                        conjunto = int(arquivo[i][63:65])
                    else:
                        mes_ent = 0
                        ano_ent = 0
                        pot_ent = 0.
                        unidade = 0
                        conjunto = 0

                    nr_expansoes += 1

                    # Trata Volume Morto
                    if mes_vm > 0:
                        volume = usina.VolMin*per_vm/100
                        volume = (usina.VolMin - volume)/dur_vm
                        mesinicial = mes_vm
                        for iano in range(ano_vm - anoinicial, nanos):
                            for imes in range(mesinicial-1, 12):
                                if dur_vm > 0:
                                    usina.StatusVolMorto[iano][imes] = 1
                                    usina.VolMortoTempo[iano][imes] = volume
                                    dur_vm -= 1
                                else:
                                    usina.StatusVolMorto[iano][imes] = 2
                                    usina.VolMortoTempo[iano][imes] = 0.
                            mesinicial = 1
                    else:
                        usina.StatusVolMorto = 2*np.ones((nanos,12),'i')

                    if mes_ent > 0:
                        mesinicial = mes_ent
                        usina.MaqporConj[conjunto-1] = unidade
                        usina.PEfporConj[conjunto-1] = pot_ent
                        usina.CalcPotEfetiva()
                        usina.CalcEngolMaximo()
                        for iano in range(ano_ent - anoinicial, nanos):
                            for imes in range(mesinicial-1, 12):
                                usina.UnidadesTempo[iano][imes] += 1
                                usina.EngolTempo[iano][imes] = usina.Engolimento
                                usina.PotenciaTempo[iano][imes] = usina.PotEfet
                            mesinicial = 1

                    i += 1
                    while arquivo[i][0:4] != '9999':
                        mes_ent = int(arquivo[i][44:46])
                        ano_ent = int(arquivo[i][47:51])
                        pot_ent = float(arquivo[i][52:58])
                        unidade = int(arquivo[i][60:62])
                        conjunto = int(arquivo[i][63:65])
                        if mes_ent > 0:
                            mesinicial = mes_ent
                            usina.MaqporConj[conjunto - 1] = unidade
                            usina.PEfporConj[conjunto - 1] = pot_ent
                            usina.CalcPotEfetiva()
                            usina.CalcEngolMaximo()
                            for iano in range(ano_ent - anoinicial, nanos):
                                for imes in range(mesinicial - 1, 12):
                                    usina.UnidadesTempo[iano][imes] += 1
                                    usina.EngolTempo[iano][imes] = usina.Engolimento
                                    usina.PotenciaTempo[iano][imes] = usina.PotEfet
                                mesinicial = 1
                        i += 1

                    for iano in range(nanos):
                        for imes in range(12):
                            if usina.UnidadesTempo[iano][imes] >= usina.UnidBase:
                                usina.StatusMotoriz[iano][imes] = 2
                            elif usina.UnidadesTempo[iano][imes] > 0:
                                usina.StatusMotoriz[iano][imes] = 1
                            else:
                                if usina.StatusVolMorto[iano][imes] == 2:
                                    usina.StatusMotoriz[iano][imes] = 1
                                else:
                                    usina.StatusMotoriz[iano][imes] = 0
                    break
            i += 1

        print('OK! Leitura do EXPH realizada com sucesso. (', nr_expansoes, 'Usinas Hidraulicas Expandidas - ', i, 'linhas lidas )')
        return(conf)

    def le_ree(self, ree):

        file_name = self.diretorio + self.nome_ree
        file = open(file_name, "r")
        arquivo = file.readlines()
        file.close()

        for i in range(3,len(arquivo)):
            if len(arquivo[i])>20:
                ree.append(resequiv())
                indice = len(ree) - 1
                ree[indice].Codigo      = int(arquivo[i][1:4])
                ree[indice].Nome        = arquivo[i][5:15]
                ree[indice].Submercado  = int(arquivo[i][18:21])
            else:
                break

        print('OK! Leitura do REE realizada com sucesso. (', indice+1, 'Reservatorios Equivalentes de Energia )' )
        return ree

    def le_sistema(self, submercado, intercambio, nanos, npmc):

        file_name = self.diretorio + self.nome_sistema
        file = open(file_name, "r")
        arquivo = file.readlines()
        file.close()

        npmc = int(arquivo[3][1:4])

        i = 7

        while arquivo[i][1:4] != '999':
            submercado.append( subsist(nanos) )
            isis = len(submercado) - 1
            submercado[isis].Codigo = int(arquivo[i][1:4])
            submercado[isis].Nome =  arquivo[i][5:15]
            submercado[isis].Ficticio = int(arquivo[i][17:18])
            if not submercado[isis].Ficticio:
                for idef in range(4):
                    submercado[isis].CustoDeficit[idef] = float(arquivo[i][19+8*idef:26+8*idef])
                    submercado[isis].ProfundidadeDeficit[idef] = float(arquivo[i][51+6*idef:56+6*idef])
            i = i + 1

        i = i + 4

        while arquivo[i][1:4] != '999':
              intercambio.append( interc(nanos) )
              iinter = len(intercambio) - 1
              intercambio[iinter].De   = int(arquivo[i][1:4])
              intercambio[iinter].Para = int(arquivo[i][5:8])
              intercambio[iinter].Flag = int(arquivo[i][23:24])

              i = i + 1

              for iano in range(nanos):
                  for imes in range(12):
                      valor = arquivo[i + iano][7 + 8 * imes:14 + 8 * imes]
                      valor = valor.strip()
                      if valor != '':
                          if intercambio[iinter].Flag == 0:
                              intercambio[iinter].LimiteMaximo[iano][imes] = float(valor)
                          else:
                              intercambio[iinter].LimiteMinimo[iano][imes] = float(valor)

              intercambio.append( interc(nanos) )
              iinter = len(intercambio) - 1
              intercambio[iinter].De   = intercambio[iinter-1].Para
              intercambio[iinter].Para = intercambio[iinter-1].De
              intercambio[iinter].Flag = intercambio[iinter-1].Flag

              i = i + nanos + 1

              for iano in range(nanos):
                  for imes in range(12):
                      valor = arquivo[i + iano][7 + 8 * imes:14 + 8 * imes]
                      valor = valor.strip()
                      if valor != '':
                          if intercambio[iinter].Flag == 0:
                              intercambio[iinter].LimiteMaximo[iano][imes] = float(valor)
                          else:
                              intercambio[iinter].LimiteMinimo[iano][imes] = float(valor)

              i = i + nanos

        i = i + 4
        while arquivo[i][1:4] != '999':
            codigo = int(arquivo[i][1:4])
            i = i + 1
            for isub in submercado:
                if isub.Codigo == codigo:
                    for iano in range(nanos+1):
                        for imes in range(12):
                            valor = arquivo[i + iano][7 + 8 * imes:14 + 8 * imes]
                            valor = valor.strip()
                            if valor != '':
                                isub.Mercado[iano][imes] = float(valor)
                    break
            i = i + nanos + 1

        i = i + 4
        while arquivo[i][1:4] != '999':
            codigo = int(arquivo[i][1:4])
            i = i + 1
            for isub in submercado:
                if isub.Codigo == codigo:
                    for iano in range(nanos):
                        for imes in range(12):
                            valor = arquivo[i + iano][7 + 8 * imes:14 + 8 * imes]
                            valor = valor.strip()
                            if valor != '':
                                isub.NaoSimuladas[iano][imes] = float(valor)
                    break
            i = i + nanos


        print('OK! Leitura do arquivo SISTEMA realizada com sucesso.')
        return submercado, intercambio, npmc







from cvxopt.modeling import matrix

class fcf(object):
    nr_cortes = None
    coef_vf = None
    termo_i = None
    estagio = None

    def __init__(self, estagio):
        self.nr_cortes = 0
        self.coef_vf = []
        self.termo_i = []
        self.estagio = estagio

    def add_corte(self, coeficientes, constante, volume, nr_estagios):
        coeficientes = -(1/nr_estagios)*coeficientes
        for i in range(0,len(volume)):
            if coeficientes[i] > 0:
                coeficientes[i] = 0
        constante = constante/nr_estagios
        for i in range(0,len(volume)):
            constante = constante - volume[i]*coeficientes[i]
        self.coef_vf.append(coeficientes)
        self.termo_i.append(constante)
        self.nr_cortes = self.nr_cortes+1

    def get_fcf(self, vf, alpha):
        rest_cortes = []
        if self.nr_cortes == 0:
            return rest_cortes
        else:
            for icor in range(0,self.nr_cortes):
                funcao = self.termo_i[icor]
                for i in range(0,len(vf)):
                    funcao = funcao + self.coef_vf[icor][i]*vf[i]
                rest_cortes.append( alpha >= funcao)
            return rest_cortes



from src import mddh
import os

sistema = mddh(os.path.join(os.getcwd(), 'pmo/'))

#i = 161
# sistema.conf_uh[i].PlotaPCV()
# sistema.conf_uh[i].PlotaPCA()
# sistema.conf_uh[i].PlotaProdutibs(0,0)
# sistema.conf_uh[i].PlotaVazoes()
# sistema.conf_uh[i].PlotaColina()
# sistema.conf_uh[i].PlotaVolume()
# sistema.conf_uh[i].PlotaVazMin()
# sistema.conf_uh[i].PlotaVolMorto()
# sistema.conf_uh[i].PlotaPotencia()
# sistema.Plota_Expansao_Uh()

# sistema.PlotaMercado()
# for i, iree in enumerate(sistema.ree):
#     sistema.ree[i].PlotaEArmMax(sistema.conf_uh)
# for i, isist in enumerate(sistema.submercado):
#     sistema.submercado[i].PlotaEArmMax(sistema.conf_uh)
#sistema.PlotaEArmMaxSist()
#sistema.PlotaEArmMaxRee()
# sistema.submercado[1].PlotaMercado()

#for ires in sistema.ree:
#    ires.PlotaEArmMax(sistema.conf_uh)
#for isis in sistema.ree:
#    isis.PlotaENA()

#for isis in sistema.submercado:
#    isis.PlotaENA()

sistema.conf_uh[7].parp(sistema.ordmaxparp)
sistema.conf_uh[7].PlotaParp(3)
sistema.conf_uh[7].gera_series_aditivo()

#for imes in range(12):
#    sistema.conf_uh[7].PlotaParp(imes)
#sistema.conf_uh[i].gera_series_aditivo()
#sistema.conf_uh[7].gera_series_multiplicativo()

#sistema.conf_uh[i].parp_otimo(6)
#sistema.conf_uh[i].pso(11)
#sistema.conf_uh[i].gera_series_multiplicativo_parp_otimo()

#sistema.pde(11, 5, 12)
#sistema.plota_corte(1,11)
#sistema.plota_corte(5,21)


